# -*- coding: utf-8 -*-
from imio.fpaudit import utils as _fpaudit_utils  # import module so it can be patched in a second time


LOG_ID = u"esign"


def audit(action, extras=""):
    """Log an eSignature action to the dedicated audit log."""
    _fpaudit_utils.fpalog(LOG_ID, action, extras)



# -*- coding: utf-8 -*-
"""Setup tests for this package."""
from imio.esign import PLONE_VERSION
from imio.esign.testing import IMIO_ESIGN_INTEGRATION_TESTING  # noqa: E501
from plone import api
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID

import unittest


if PLONE_VERSION >= 5:
    from Products.CMFPlone.utils import get_installer


class TestSetup(unittest.TestCase):
    """Test that imio.esign is properly installed."""

    layer = IMIO_ESIGN_INTEGRATION_TESTING

    def setUp(self):
        """Custom shared utility setup for tests."""
        self.portal = self.layer["portal"]
        if PLONE_VERSION < 5:
            self.installer = api.portal.get_tool("portal_quickinstaller")
        else:
            self.installer = get_installer(self.portal, self.layer["request"])

    def test_product_installed(self):
        """Test if imio.esign is installed."""
        if PLONE_VERSION < 5:
            self.assertTrue(self.installer.isProductInstalled("imio.esign"))
        else:
            self.assertTrue(self.installer.is_product_installed("imio.esign"))

    def test_browserlayer(self):
        """Test that IImioEsignLayer is registered."""
        from imio.esign.interfaces import IImioEsignLayer
        from plone.browserlayer import utils

        self.assertIn(IImioEsignLayer, utils.registered_layers())


class TestUninstall(unittest.TestCase):

    layer = IMIO_ESIGN_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        roles_before = api.user.get_roles(TEST_USER_ID)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        if PLONE_VERSION < 5:
            self.installer = api.portal.get_tool("portal_quickinstaller")
            self.installer.uninstallProducts(["imio.esign"])
        else:
            self.installer = get_installer(self.portal, self.layer["request"])
            self.installer.uninstall_product("imio.esign")
        setRoles(self.portal, TEST_USER_ID, roles_before)

    def test_product_uninstalled(self):
        """Test if imio.esign is cleanly uninstalled."""
        if PLONE_VERSION < 5:
            self.assertFalse(self.installer.isProductInstalled("imio.esign"))
        else:
            self.assertFalse(self.installer.is_product_installed("imio.esign"))

    def test_browserlayer_removed(self):
        """Test that IImioEsignLayer is removed."""
        from imio.esign.interfaces import IImioEsignLayer
        from plone.browserlayer import utils

        self.assertNotIn(IImioEsignLayer, utils.registered_layers())



# -*- coding: utf-8 -*-
"""Init and utils."""
from plone import api
from zope.component import queryUtility
from zope.i18n import ITranslationDomain
from zope.i18nmessageid import MessageFactory

import logging
import os


_ = MessageFactory("imio.esign")
logger = logging.getLogger("imio.esign")
PLONE_VERSION = int(api.env.plone_version()[0])
API_ROOT_URL = os.getenv("API_ROOT_URL", "http://127.0.0.1:8000")
manage_session_perm = "imio.esign: Manage Sessions"


def _tr(msgid, domain="imio.esign", mapping=None):
    translation_domain = queryUtility(ITranslationDomain, domain)
    return translation_domain.translate(msgid, target_language=api.portal.get_current_language(), mapping=mapping)



# -*- coding: utf-8 -*-
"""Module where all interfaces, events and exceptions live."""

from zope.interface import Interface
from zope.publisher.interfaces.browser import IDefaultBrowserLayer


class IImioEsignLayer(IDefaultBrowserLayer):
    """Marker interface that defines a browser layer."""


class IContextUidProvider(Interface):
    """Adapter to provide context UID for a file."""

    def get_context_uid():
        """Return the context UID for the file to sign."""



# -*- coding: utf-8 -*-

from eea.facetednavigation.interfaces import IFacetedNavigable
from html import escape
from imio.esign import _
from imio.esign.config import get_esign_registry_seal_code
from imio.esign.config import get_esign_registry_seal_email
from imio.esign.utils import get_state_description
from imio.helpers.security import check_zope_admin
from imio.pyutils.utils import safe_encode
from plone import api
from Products.CMFPlone.utils import safe_unicode
from z3c.table.column import Column
from z3c.table.table import Table
from zope.component import getMultiAdapter
from zope.i18n import translate


class IdColumn(Column):
    # not translated so it stays short
    header = "Id"
    weight = 10
    cssClasses = {"th": "th_header_sessions_id",
                  "td": "id-column"}

    def renderHeadCell(self):
        """
        Override to add JS that will the "No results" link when displayed in faceted dashboard.
        """
        res = super(IdColumn, self).renderHeadCell()
        if IFacetedNavigable.providedBy(self.context):
            res += "<script>$('div.table_faceted_results').hide();</script>"
        return res

    def renderCell(self, item):
        # this will hide the "No results" link when displayed in faceted dashboard
        return "<span id='{0}'>{0}</span>".format(str(item.get("id")))


def external_session_link(session, title=None):
    """Return a tag with the sign external session."""
    title = title or session.get("title", "") or session.get("sign_id", "")
    if not session["sign_id"] or not session["sign_url"]:
        return u"<span>{0}</span>".format(title)
    return u'<a href="{url}" target="_blank">{title}</a>'.format(
        url=session["sign_url"],
        title=safe_unicode(title),
    )


class StateColumn(Column):
    header = _("State")
    weight = 20
    cssClasses = {"th": "th_header_sessions_state",
                  "td": "state-column"}

    def renderCell(self, item):
        state = escape(translate(
            (item.get("state", "")), context=self.request, default=item.get("state", ""), domain="imio.esign",
        ))
        title = escape(translate(get_state_description(item.get("state", "")), context=self.request,
                                 domain="imio.esign"))
        return (u"<span class='state-title state-title-{state_title_value}' title='{title}'>{state} <span class='far fa-question-circle' />"
                u"</span>".format(state=state, title=title, state_title_value=item.get("state")))


class TitleColumn(Column):
    header = _("Title")
    weight = 30
    cssClasses = {"th": "th_header_sessions_title",
                  "td": "title-column"}

    def renderCell(self, item):
        title = safe_unicode(item.get("title", ""))
        if item["sign_url"]:
            return external_session_link(item, title=title)
        else:
            return title


class SealColumn(Column):
    header = _("Sealed")
    weight = 40
    cssClasses = {"th": "th_header_sessions_seal nosort",
                  "td": "seal-column"}

    def renderCell(self, item):
        if not item.get("seal"):
            return u""
        label = escape(translate(_("Sealed"), context=self.request))
        # icon: https://www.flaticon.com/free-icon/verification_3556787
        return (u"<img width='16' height='16' src='++resource++imio.esign/seal.png' title='{label}' "
                u"aria-label='{label}'></img>".format(label=label))


class SignersColumn(Column):
    header = _("Signers")
    weight = 50
    cssClasses = {"th": "th_header_sessions_signers",
                  "td": "signers-column"}

    def renderCell(self, item):
        signers = item.get("signers") or []
        parts = [
            "<li>%s, %s%s (%s)</li>" % (
                safe_encode(s.get("fullname", "")),
                safe_encode(s.get("position")),
                " (%s)" % safe_encode(translate(s.get("status"), domain="imio.esign", context=self.request)) if s.get("status") else "",
                s.get("email"), )
            for s in signers
        ]
        return safe_unicode("<ol>%s</ol>" % "".join(parts))


class FilesColumn(Column):
    header = _("Files")
    weight = 60
    cssClasses = {"th": "th_header_sessions_documents nosort",
                  "td": "documents-column"}

    def renderCell(self, item):
        """Render a collapsible block that loads the list on demand."""
        # Row identifier (unique per session)
        session_id = item.get("id")
        details_msg = translate(_("Quick look"), context=self.request)
        base_url = getattr(self.table, "portal_url", None)
        if not base_url:
            try:
                base_url = self.context.absolute_url()
            except Exception:
                base_url = api.portal.get().absolute_url()

        html = (
            u'<div id="session-files" class="collapsible" '
            u"onclick=\"toggleDetails('collapsible-session-files_{0}', "
            u"toggle_parent_active=true, parent_tag=null, "
            u"load_view='@@esign-session-files?session_id={0}', "
            u"base_url='{1}');\"> {2}</div>"
            u'<div id="collapsible-session-files_{0}" class="collapsible-content" style="display: none;">'
            u'<div class="collapsible-inner-content">'
            u'<img src="{1}/spinner_small.gif" />'
            u"</div></div>"
        ).format(session_id, base_url, details_msg)

        return html


class LastUpdateColumn(Column):
    header = _("Last update")
    weight = 70
    cssClasses = {"th": "th_header_sessions_last_update",
                  "td": "last-update-column"}

    def renderCell(self, item):
        last_update = item.get("last_update")
        # make sortable
        value = "<span style='display:none'>{0}</span>".format(last_update)
        return value + self.context.unrestrictedTraverse('@@plone').toLocalizedTime(
            last_update, long_format=True)


class ActionsColumn(Column):
    """ """

    header = _("Actions")
    weight = 80
    cssClasses = {"th": "th_header_sessions_actions nosort",
                  "td": "actions-column"}

    def renderCell(self, item):
        session_id = item.get("id")
        dashboard_link = self.table.view.get_dashboard_link({"id": session_id})
        sessions_url = self.table.view.get_sessions_url()
        portal = api.portal.get()
        # if not sessions_url.endswith("/"):
        #    sessions_url += "/"
        admin_buttons = u""
        if getMultiAdapter((portal, self.request), name="esign-session-delete").may_delete_session():
            admin_buttons = u"""
            <img width="16" height="16" title="{delete}" style="cursor:pointer" src="delete_icon.png"
            onclick="javascript:confirmDeleteObject(base_url='{sessions_url}', object_uid=null, this,
            msgName=null, view_name='@@esign-session-delete?esign_session_id={session_id}', redirect=null);">
            """.format(
                delete=translate(_("Delete session"), context=self.request),
                sessions_url=sessions_url,
                session_id=session_id,
            )
        if (item.get("state") == "draft"
                and getMultiAdapter((portal, self.request),
                                    name="external-esign-session-create").may_create_external_sessions()):
            admin_buttons += u"""
            <img width="16" height="16" title="{send}" style="cursor:pointer" src="++resource++imio.esign/parapheo.svg"
            onclick="javascript:callViewAndReload('{sessions_url}','@@external-esign-session-create',
            {{'session_id': '{session_id}'}});">
            """.format(
                sessions_url=sessions_url,
                session_id=session_id,
                send=translate(_("Create external session"), context=self.request),
            )
        if check_zope_admin():
            admin_buttons += u"""
            <a class="link-overlay-info" href="{sessions_url}/@@session-annotation-info?session_id={session_id}" target="_blank">
                <span class="fa fa-info-circle" title="Annotation info"></span>
            </a>
            """.format(
                sessions_url=sessions_url,
                session_id=session_id,
            )
        dashboard_button = u"""
        <a href="{dashboard_link}"><img title="{dashboard_view}" style="cursor:pointer"
        src="++resource++imio.esign/view_element.png"></a>
        """.format(  # noqa E501
            dashboard_link=dashboard_link,
            dashboard_view=translate(_("View session in dashboard"), context=self.request),
        )
        return admin_buttons + dashboard_button


class SessionsTable(Table):
    cssClassEven = "even"
    cssClassOdd = "odd"
    cssClasses = {"table": "listing sessions-table width-full"}
    sortOn = None
    results = []

    def __init__(self, context, view, request, items=None):
        super(SessionsTable, self).__init__(context, request)
        self.view = view
        self._items = items

    @property
    def values(self):
        return self._items

    def setUpColumns(self):
        ctx, req, tbl = self.context, self.request, self
        columns = [
            IdColumn(ctx, req, tbl),
            StateColumn(ctx, req, tbl),
            TitleColumn(ctx, req, tbl),
            SignersColumn(ctx, req, tbl),
            FilesColumn(ctx, req, tbl),
            LastUpdateColumn(ctx, req, tbl),
            ActionsColumn(ctx, req, tbl),
        ]
        if get_esign_registry_seal_code() and get_esign_registry_seal_email():
            seal_col = SealColumn(ctx, req, tbl)
            columns.insert(4, seal_col)
        return columns



# -*- coding: utf-8 -*-
"""
Create a technical user in Luxtrust and upload an image.

This script allows to:
- check for existing users in the Luxtrust directory
- create a new user if it does not exist
- upload an acroform image for the user

Created from https://gitlab.imio.be/imio-api/apims-esign/-/blob/se-221-create_users_update_watermark-jean/notebooks/
COSI_eSign_put_acroform.ipynb
"""
from imio.pyutils.system import post_request
from imio.pyutils.system import stop

import argparse
import logging
import os.path
import requests


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("cosi_push")


def get_bearer_token(cosi_url, directory_alias, tenant_name, username, password):
    """Authentication and Bearer token get."""
    url = "{}/auth?directoryAlias={}".format(cosi_url, directory_alias)
    data = {
        "tenantName": tenant_name,
        "username": username,
        "password": password,
    }
    response = post_request(url, json=data, logger=logger, return_json=True)
    if not response:
        stop("No response from auth request", logger=logger)
    token = response.get("access_token")  # noqa F841
    if not token:
        stop("No token in auth response", logger=logger)
    return token


def get_users_from_directory(cosi_url, directory_alias, users, headers):
    """Get cosi users from email list."""
    payload = {
        "directoryAlias": directory_alias,
        "userIdList": users,
    }

    user_ids = post_request(url="{}/api/endusers/search".format(cosi_url), data=payload, headers=headers,
                            return_json=True)
    if not user_ids:
        return []
    for user in user_ids:
        logger.info("  - {}: {} ({})".format(user.get('id'), user.get('userId'), user.get('email')))
    return user_ids


def create_new_user(cosi_url, directory_alias, user, organisation_id, headers):
    """Create user."""
    firstname, lastname = user.split("@")
    payload = {
        "directoryAlias": directory_alias,
        "userId": user,
        "email": user,
        "firstName": firstname,
        "lastName": lastname,
        "localeCode": "FR",
        "phone": "+3200000000",
        "organisationId": organisation_id,
    }
    new_user = post_request(url="{}/api/endusers".format(cosi_url), json=payload, headers=headers)
    if new_user and new_user.get("id"):
        user_id = new_user.get("id")
        logger.info("New user created with id {}, email {}".format(user_id, user))
        return user_id
    else:
        stop("Cannot create user {}".format(user), logger=logger)
        return None


def upload_acroform(cosi_url, user_id, file_path, headers):
    """Upload image from path."""
    # 1. Get watermark ID for the user
    response = requests.get(url="{}/api/watermarks?targetId={}&type=USER".format(cosi_url, user_id), headers=headers)
    if not response.ok:
        stop("Error getting watermark for user {}".format(user_id), logger=logger)
    watermark_response = response.json()
    if not watermark_response or not isinstance(watermark_response, list) or len(watermark_response) == 0:
        stop("Error getting watermark for user {}".format(user_id), logger=logger)

    watermark_id = watermark_response[0].get("id")
    logger.debug("Watermark ID: {}".format(watermark_id))

    # 2. Upload image file
    with open(file_path, "rb") as f:
        file_content = f.read()

    filename = os.path.basename(file_path)
    multipart_data = [("image", (filename, file_content, "image"))]
    url = "{}/api/watermarks/{}?targetId={}&type=USER".format(cosi_url, watermark_id, user_id)

    result = requests.put(url=url, files=multipart_data, headers=headers)

    if result.ok:
        logger.info("Image uploaded for user {}".format(user_id))
    else:
        logger.error("Error uploading image for user {}: {}".format(user_id, result.text))


def main():
    """Fonction principale."""
    parser = argparse.ArgumentParser(description="Create user in Luxtrust and upload acroform image.")

    parser.add_argument("username", help="COSI username")
    parser.add_argument("password", help="COSI password")
    parser.add_argument("user_email", help="User email to check/create in Luxtrust")

    parser.add_argument("-f", "--file", help="File path to upload")
    parser.add_argument("-s", "--simulation", action="store_true", help="Simulation mode")
    parser.add_argument("--tenant-name", default="IMIO", help="Tenant name (default: IMIO)")
    parser.add_argument("--directory-alias", default="imio_internal", help="Directory alias (default: imio_internal)")
    parser.add_argument("--organisation-id", type=int, default=70, help="Organisation ID (default: 70 for staging)")
    parser.add_argument("--cosi-url", default="https://simplycosi-1-test.trustsigneurope.com", help="COSI URL")

    args = parser.parse_args()

    token = get_bearer_token(args.cosi_url, args.directory_alias, args.tenant_name, args.username, args.password)
    headers = {"Authorization": "Bearer {}".format(token)}

    logger.info("Checking if user '{}' exists...".format(args.user_email))
    user_ids = get_users_from_directory(args.cosi_url, args.directory_alias, [args.user_email], headers)

    user_id = None
    for u in user_ids:
        if u.get("userId") == args.user_email:
            user_id = u.get("id")
            logger.info("User '{}' found. We will reuse it".format(args.user_email))
            break

    if not user_id:
        logger.info("User '{}' not found. We will create it".format(args.user_email))
        if args.simulation:
            logger.info("Simulation mode: skipping user creation.")
        else:
            user_id = create_new_user(
                args.cosi_url,
                args.directory_alias,
                args.user_email,
                args.organisation_id,
                headers,
            )

    if args.file:
        logger.info("Upload image '{}'".format(args.file))
        if args.simulation:
            logger.info("Simulation mode: skipping image upload.")
        else:
            upload_acroform(args.cosi_url, user_id, args.file, headers)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from plone import api


def get_esign_registry_enabled(default=False):
    return api.portal.get_registry_record("imio.esign.enabled", default=default)


def get_esign_registry_vat_number(default=""):
    return api.portal.get_registry_record("imio.esign.vat_number", default=default)


def get_esign_registry_file_url(default=""):
    return api.portal.get_registry_record("imio.esign.file_url", default=default)


def get_esign_registry_seal_code(default=""):
    return api.portal.get_registry_record("imio.esign.seal_code", default=default)


def get_esign_registry_seal_email(default=""):
    return api.portal.get_registry_record("imio.esign.seal_email", default=default)


def get_esign_registry_sign_code(default=""):
    return api.portal.get_registry_record("imio.esign.sign_code", default=default)


def get_esign_registry_parapheo_url(default=""):
    return api.portal.get_registry_record("imio.esign.parapheo_url", default=default)


def get_esign_registry_signing_users_email_content(default=""):
    return api.portal.get_registry_record("imio.esign.signing_users_email_content", default=default)


def get_esign_registry_max_session_size(default=100):
    return api.portal.get_registry_record("imio.esign.max_session_size", default=default)


def get_esign_registry_external_watchers():
    value = api.portal.get_registry_record("imio.esign.external_watchers", default="")
    if not value:
        return []
    return [ew.strip() for ew in value.split(",") if ew.strip()]


def set_esign_registry_enabled(value):
    api.portal.set_registry_record("imio.esign.enabled", value)


def set_esign_registry_vat_number(value):
    api.portal.set_registry_record("imio.esign.vat_number", value)


def set_esign_registry_file_url(value):
    api.portal.set_registry_record("imio.esign.file_url", value)


def set_esign_registry_seal_code(value):
    api.portal.set_registry_record("imio.esign.seal_code", value)


def set_esign_registry_seal_email(value):
    api.portal.set_registry_record("imio.esign.seal_email", value)


def set_esign_registry_sign_code(value):
    api.portal.set_registry_record("imio.esign.sign_code", value)


def set_esign_registry_parapheo_url(value):
    api.portal.set_registry_record("imio.esign.parapheo_url", value)


def set_esign_registry_signing_users_email_content(value):
    api.portal.set_registry_record("imio.esign.signing_users_email_content", value)


def set_esign_registry_max_session_size(value):
    api.portal.set_registry_record("imio.esign.max_session_size", value)


def set_esign_registry_external_watchers(value):
    api.portal.set_registry_record("imio.esign.external_watchers", value)


SIGNERS_EMAIL_CONTENT = u"""
<meta charset="UTF-8"><tal:global>
<p style="font-weight: bold;" tal:condition="nothing">!! Attention: ne pas modifier ceci directement mais passer
par "Source" !!</p>

<p>Bonjour <span tal:content="python: user_data['fullname']">FULLNAME</span>,</p>

<p>Vous avez été défini comme signataire dans une application IMIO (iA.Delib, iA.Docs, etc.).
<br />Avant de pouvoir signer des documents, vous devez activer votre compte auprès de Paraphéo.</p>

<p>Pour ce faire, il est nécessaire de s'y connecter une toute première fois en suivant ces étapes:</p>

<ol>
<li>Vous rendre sur <a href="#" tal:attributes="href parapheo_url">Paraphéo</a></li>
<li>Cliquer sur le mode de connexion "Portal d'authentification"</li>
<li>Entrer votre adresse email "<span tal:content="python: user_data['email']">EMAIL</span>" et cliquer
sur "Connexion"</li>
</ol>

<p>Si vous n'avez jamais défini votre mot de passe dans Wallonie Connect ou que vous l'avez oublié:</p>

<ol>
<li>Cliquer sur "Mot de passe oublié ?" (situé sous le champ "mot de passe")</li>
<li>Entrer à nouveau votre email et cliquer sur "Soumettre"</li>
<li>Consulter votre boîte mail qui doit contenir un mail intitulé "Réinitialiser le mot de passe"
(vérifier les spams au cas ou...)</li>
<li>Suivre les étapes pour configurer l'authentification par mobile</li>
</ol>

<p>En cas de besoin, n'hésitez pas à faire appel à votre référent interne Wallonie Connect !</p>

<p>Une fois votre mot de passe défini, vous pouvez poursuivre l'identification sur le site Paraphéo:</p>

<ol>
<li>Si besoin, répéter les 3 premières étapes du premier paragraphe</li>
<li>Entrer votre adresse email et&nbsp;votre mot de passe et cliquer sur "Connexion"</li>
<li>Entrer votre "code à usage unique" depuis votre mobile et cliquer sur "Connexion"</li>
<li>Une fois connecté dans Paraphéo, votre compte est validé et prêt à être utilisé dans une session de signature.</li>
</ol>

<p>Cordialement
<br />L'équipe IMIO</p>
</tal:global>"""



# -*- coding: utf-8 -*-

from imio.esign.utils import get_file_info
from imio.esign.utils import get_sessions_for
from imio.helpers.transmogrifier import get_correct_id
from os import path


def on_categorized_annex_updated(annex, event):
    '''When an annex is modified, update check if need to update esign session.'''
    old_values = event.old_values
    # we are creating a new annex, not in a session
    if not old_values:
        return

    sessions = get_sessions_for(event.parent.UID(), readonly=False)
    if not sessions:
        return

    # make sure annex_uid is in a session
    annex_uid = annex.UID()
    file_infos = []
    for session_id in sessions:
        file_info = get_file_info(session_id, annex_uid)
        if file_info:
            file_infos.append(file_info)
    if not file_infos:
        return

    # here we are sure that annex is in a session, we need to update data
    # if something usefull changed, we will update the session
    new_values = event.new_values
    update = False
    checked_keys = ['title', 'filesize', 'relative_url']
    for checked_key in checked_keys:
        if new_values[checked_key] != old_values[checked_key]:
            update = True
            break
    # check scan_id and filename
    if update is False:
        for file_info in file_infos:
            if file_info and (annex.scan_id != file_info['scan_id'] or \
               annex.file.filename != file_info['filename']):
                update = True
                break

    if update is True:
        for session_id, session in sessions.items():
            # size
            size_diff = new_values['filesize'] - old_values['filesize']
            session['size'] += size_diff
            # title and filename
            for file_data in session['files']:
                if file_data['uid'] == annex_uid:
                    file_data['title'] = new_values['title']
                    file_data['scan_id'] = annex.scan_id
                    # filename changed, need to make sure new filename is unique
                    if annex.file.filename != file_data['filename']:
                        existing_files = [path.splitext(f["filename"])[0]
                                          for f in session["files"]]
                        filename, ext = path.splitext(annex.file.filename)
                        new_filename = get_correct_id(existing_files, filename)
                        file_data['filename'] = new_filename + ext
                    # file_uid is only there one time per session
                    break



# -*- coding: utf-8 -*-

from imio.esign.utils import get_session_annotation
from imio.esign.utils import get_session_info
from plone.memoize import ram
from zope.interface import Interface


class DefaultContextUidProvider(object):
    """Default adapter that returns parent UID as context."""

    def __init__(self, context):
        self.context = context

    def get_context_uid(self):
        """Return parent UID as default context."""
        parent = getattr(self.context, "aq_parent", None)
        if parent and hasattr(parent, "UID"):
            return parent.UID()
        return None


class FilesBelongingToAGivenSession(object):
    def __init__(self, context):
        self.context = context
        self.request = self.context.REQUEST

    def query_session_files_cachekey(method, self):
        return (get_session_annotation()._p_mtime, self.get_session_id())

    @property
    @ram.cache(query_session_files_cachekey)
    def query_session_files(self):
        session_id = self.get_session_id()
        if not session_id:
            return {"UID": {"query": []}}
        session = get_session_info(int(session_id))
        obj_uids = []
        for f in session.get("files", []):
            obj_uids.append(f["context_uid"])
        return {"UID": {"query": obj_uids}}

    def get_session_id(self):
        if "esign_session_id[]" in self.request.form.keys():
            return self.request.form["esign_session_id[]"]
        else:
            return self.request.get("esign_session_id", None)

    query = query_session_files


class ISignable(Interface):
    def get_signers(self):
        """
        List signers for that element.
        returns a list of dict with keys:
        name, function, held_position, userid and email
        Ensure a valid userid is linked to the held_position.person and a
        unique email is used for every signers
        Raise ValueError in case an error occurs.
        """
        return []

    def get_files_uids(self):
        """
        List of file uids to sign.
        """
        return []

    def get_watchers(self):
        """
        List of watchers email for that element.
        """
        return []

    def get_discriminators(self):
        """
        List of discriminators for that element to associate it to an available session.
        """
        return []

    def get_create_session_custom_data(self):
        """
        Dict of custom data that will be stored on session at creation time.
        """
        return {}



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from datetime import datetime
from datetime import timedelta
from imio.esign import _
from imio.esign import manage_session_perm
from imio.esign.audit import audit
from imio.esign.browser.table import external_session_link
from imio.esign.browser.table import SessionsTable
from imio.esign.config import get_esign_registry_enabled
from imio.esign.config import get_esign_registry_parapheo_url
from imio.esign.config import get_esign_registry_signing_users_email_content
from imio.esign.utils import create_external_session
from imio.esign.utils import get_session_annotation
from imio.esign.utils import get_session_info
from imio.esign.utils import get_sessions_for
from imio.esign.utils import get_state_description
from imio.esign.utils import remove_session
from imio.helpers.content import uuidToObject
from imio.helpers.emailer import create_html_email
from imio.helpers.emailer import send_email
from imio.helpers.security import separate_fullname
from imio.prettylink.interfaces import IPrettyLink
from imio.pyutils.utils import safe_encode
from imio.pyutils.utils import shortuid_decode_id
from plone import api
from plone.app.layout.viewlets import ViewletBase
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import base_hasattr
from Products.Five import BrowserView
from Products.PageTemplates.Expressions import SecureModuleImporter
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import CachedProperty
from zope.component import getMultiAdapter
from zope.i18n import translate
from zope.interface import implementer
from zope.pagetemplate.pagetemplate import PageTemplate
from zope.publisher.interfaces import IPublishTraverse

import csv
import json
import os


try:
    from StringIO import StringIO  # Python 2
except ImportError:
    from io import StringIO  # Python 3


class SessionsListingView(BrowserView):
    """View to list sessions."""

    def __init__(self, context, request):
        super(SessionsListingView, self).__init__(context, request)
        self.portal = api.portal.get()
        self.portal_url = self.portal.absolute_url()

    def __call__(self):
        if not self.available():
            raise Unauthorized
        return super(SessionsListingView, self).__call__()

    def available(self):
        return get_esign_registry_enabled()

    def render_table(self):
        table = SessionsTable(self.context, self, self.request, self.get_sessions())
        table.update()
        return table.render()

    def get_sessions(self):
        sessions = []
        for session_id, session in sorted(get_session_annotation()["sessions"].items(), key=lambda x: x[0],
                                          reverse=True):
            session["id"] = session_id
            sessions.append(session)
        return sessions

    def get_dashboard_link(self, session):
        raise NotImplementedError

    def get_sessions_url(self):
        return self.portal_url


class SessionFilesView(BrowserView):
    """View to display documents of a session."""

    def __init__(self, context, request):
        super(SessionFilesView, self).__init__(context, request)
        self.files = []

    def __call__(self):
        session_id = int(self.request.get("session_id"))
        session = self.get_session(session_id)
        files = []
        for f in session["files"]:
            ctx = uuidToObject(f["context_uid"])
            obj = uuidToObject(f["uid"])
            if obj and ctx:
                files.append((ctx, obj))
        self.files = files
        return self.index()

    def get_session(self, session_id):
        """Get the session object."""
        return get_session_annotation()["sessions"][session_id]

    def get_file_link(self, ctx, obj):
        return IPrettyLink(ctx).getLink() + " / " + IPrettyLink(obj).getLink()


class SessionDeleteView(BrowserView):
    """View to delete a session."""

    def __call__(self):
        if not self.may_delete_session():
            raise Unauthorized
        session_id = self.request.get("esign_session_id")
        if not session_id:
            api.portal.show_message(_("No session ID provided!"), request=self.request, type="error")
            return self.request.RESPONSE.redirect(self.context.absolute_url())

        session_id = int(session_id)
        sessions = get_session_annotation()["sessions"]
        if session_id in sessions:
            remove_session(session_id)
            audit("delete_session", "session={}".format(session_id))
            api.portal.show_message(_("Session deleted successfully!"), request=self.request, type="info")
        else:
            api.portal.show_message(_("Session not found!"), request=self.request, type="error")

        return self.request.RESPONSE.redirect(self.context.absolute_url() + "/@@parapheo")

    def may_delete_session(self):
        """Check if the user may delete sessions"""
        return api.user.has_permission(manage_session_perm, obj=self.context)


class ExternalSessionCreateView(BrowserView):
    """View to create a session in Luxtrust."""

    def __call__(self, session_id=None):
        if not self.may_create_external_sessions():
            raise Unauthorized
        if session_id is None:
            session_id = self.request.get("session_id", None)
        if session_id is None:
            api.portal.show_message(_("No session ID provided!"), request=self.request, type="error")
            return self.context.absolute_url() + "/@@parapheo"
        resp = create_external_session(int(session_id))
        if resp == "_session_not_found_":
            audit("send_to_external_service", "session={} error=session_not_found".format(session_id))
            api.portal.show_message(
                _("Session with ID ${id} doesn't exist anymore !", mapping={"id": session_id}),
                request=self.request,
                type="error",
            )
        elif resp == "_no_seal_code_":
            audit("send_to_external_service", "session={} error=no_seal_code".format(session_id))
            api.portal.show_message(
                _("No seal code defined in configuration ! Session ${id} not sent.", mapping={"id": session_id}),
                request=self.request,
                type="error",
            )
        elif resp == "_no_seal_email_":
            audit("send_to_external_service", "session={} error=no_seal_email".format(session_id))
            api.portal.show_message(
                _("No seal email defined in configuration ! Session ${id} not sent.", mapping={"id": session_id}),
                request=self.request,
                type="error",
            )
        elif resp == "_no_files_":
            audit("send_to_external_service", "session={} error=no_files".format(session_id))
            api.portal.show_message(
                _("No files found to be sent ! Session ${id} not sent.", mapping={"id": session_id}),
                request=self.request,
                type="error",
            )
        elif resp.status_code == 200:
            audit("send_to_external_service", "session={} status=200".format(session_id))
            api.portal.show_message(_("External session sent successfully!"), request=self.request, type="info")
        else:
            audit("send_to_external_service", "session={} status={} reason={}".format(
                session_id, resp.status_code, resp.reason))
            api.portal.show_message(
                _("Error while sending session: ${error}", mapping={"error": "{} {} {}".format(
                    resp.status_code, resp.reason, resp.text)}),
                request=self.request,
                type="error",
            )
        return self.context.absolute_url() + "/@@parapheo"

    def may_create_external_sessions(self):
        """Check if the user may create external sessions"""
        return api.user.has_permission(manage_session_perm, obj=self.context)


class FacetedSessionInfoViewlet(ViewletBase):
    """Show selected session info inside faceted results."""

    sessions_listing_view = SessionsListingView  # to be overridden in subclass

    def available(self):
        """Global availability of the viewlet."""
        if self.sessions_collection_uid is None:
            return False
        return True

    @property
    def sessions_collection_uid(self):
        raise NotImplementedError("You must set sessions_collection_uid in subclass.")

    def render(self):
        """Render the viewlet."""
        if self.request.form.get("c1[]", None) == self.sessions_collection_uid:
            if self.sessions:
                return self.index()
            return self.sessions_listing_view(self.context, self.request).render_table()
        return ""

    @CachedProperty
    def sessions(self):
        session_id = self.request.form.get("esign_session_id[]", None)
        try:
            session_id = int(session_id)
        except (TypeError, ValueError):
            return {}
        session = {}
        session_info = get_session_info(session_id)
        if session_info:
            session = {session_id: session_info}
        return session

    def get_table_rows(self, column):
        """Get the table rows following the column"""
        return {1: ["session_id", "state", "update_date", "sealed"],
                2: ["external_link", "signers"]}.get(column, [])

    def ext_session_link(self, session):
        return external_session_link(session)

    @property
    def session_listing_url(self):
        return api.portal.get().absolute_url() + "/@@parapheo"

    def can_display_sessions_listing_link(self):
        return getMultiAdapter((api.portal.get(), self.request), name="parapheo").available()

    def collapsible_css_default(self):
        """Default CSS class to apply on the collapsible."""
        return "collapsible active"

    def collapsible_content_css_default(self):
        """Default CSS class to apply on the collapsible."""
        return "collapsible-content"

    def get_state_description(self, state):
        return translate(get_state_description(state), context=self.request, domain="imio.esign")


class ItemSessionInfoViewlet(FacetedSessionInfoViewlet):
    """Show session info for all sessions linked to a context item."""

    def available(self):
        """Global availability of the viewlet."""
        return True

    def render(self):
        """Render the viewlet."""
        if self.sessions:
            return self.index()
        return ""

    @CachedProperty
    def sessions(self):
        """Return all sessions that contain files from this context."""
        return get_sessions_for(self.context.UID())


@implementer(IPublishTraverse)
class DownloadFileView(BrowserView):
    """View to download a file based on an identifier passed in the URL path.

    Finally not used !"""

    shortuid_separator = "-"
    named_blob_file_attribute = "file"
    download_time_delta = timedelta(days=120)

    def __init__(self, context, request):
        super(DownloadFileView, self).__init__(context, request)
        self.file_id = None

    def publishTraverse(self, request, name):
        """Capture the file identifier from the URL path.

        This method is called by Zope's traversal mechanism when accessing
        /download-file/1234-567. It captures '1234-567'.
        """
        if self.file_id is None:
            self.file_id = name
        else:
            pass
        return self

    def __call__(self):
        """Handle the file download request and return a html response."""
        if self.file_id is None:
            message = translate(_("A file identifier must be passed in the url !"), context=self.request)
            return self.html_message(message)
        decoded_uid = shortuid_decode_id(self.file_id, self.shortuid_separator)
        if decoded_uid is None:
            message = translate(_("This file identifier is not correct !"), context=self.request)
            return self.html_message(message)
        file_obj = uuidToObject(decoded_uid, unrestricted=True)
        if file_obj is None:
            message = translate(_("The corresponding file identifier cannot be retrieved (${uid}) !",
                                  mapping={"uid": safe_encode(self.file_id)}),
                                context=self.request)
            return self.html_message(message)
        # Verify date - check if file is not too old
        if self.download_time_delta is not None:
            modification_date = file_obj.modified()
            if hasattr(modification_date, 'asdatetime'):
                modification_date = modification_date.asdatetime()
            modification_date = modification_date.date()
            if datetime.now().date() - modification_date > self.download_time_delta:
                message = translate(
                    _("The download period for this file has expired (was ${valid_date}) !",
                      mapping={"valid_date": datetime.strftime(modification_date + self.download_time_delta,
                                                               "%Y-%m-%d")}),
                    context=self.request)
                return self.html_message(message)
        # Get file content
        nbf = getattr(file_obj, self.named_blob_file_attribute, None)
        if nbf is None:
            message = translate(_("The corresponding file content cannot be retrieved (${uid}) !",
                                  mapping={"uid": safe_encode(decoded_uid)}),
                                context=self.request)
            return self.html_message(message)
        # Serve the file
        response = self.request.RESPONSE
        filename = safe_encode(nbf.filename)
        if "__" in filename:
            filename = filename.split("__")[0] + os.path.splitext(filename)[1]
        response.setHeader("Content-Type", nbf.contentType)
        response.setHeader("Content-Disposition", 'inline; filename="{}"'.format(filename))
        response.setHeader("Content-Length", str(len(nbf.data)))
        return nbf.data

    def html_message(self, message):
        """Returns a html message

        :param message: translated message to display
        :return: File content or HTML response
        """
        response = self.request.RESPONSE
        response.setHeader('Content-Type', 'text/html; charset=utf-8')

        # Translate HTML content
        page_title = translate(_("Signed file download"), context=self.request)
        heading = translate(_("Signed file download"), context=self.request)

        html = u"""
        <!DOCTYPE html>
        <html>
        <head>
            <meta charset='utf-8'>
            <title>{title}</title>
            <style>
                body {{ font-family: Arial, sans-serif; margin: 40px; }}
                .info-box {{
                    background-color: #fff3cd;
                    border: 2px solid #ff9800;
                    padding: 20px;
                    border-radius: 5px;
                }}
                h1 {{ color: #e65100; }}
                p {{ color: #663c00; }}
            </style>
        </head>
        <body>
            <div class='info-box'>
                <h1>⚠️ {heading}</h1>
                <p>{message}</p>
            </div>
        </body>
        </html>
        """.format(
            title=page_title, heading=heading, message=message
        )

        return html


class SigningUsersCsv(BrowserView):
    """Get users, checking for duplicate emails, output a CSV, and send emails.
    This view can be subclassed to redefine custom filtering logic.
    """

    index = ViewPageTemplateFile("templates/signing_users.pt")

    def __call__(self):
        # Handle CSV download
        if self.request.get("action") == "download_csv":
            return self._download_csv()

        # Handle email sending
        if self.request.get("action") == "send_emails":
            return self._send_emails()

        # Default: display the table
        return self.index()

    def filter_user(self, user_data):
        """Filter method to determine if a user should be included by default.

        :param user_data: dict containing user data (userid, email, lastname, firstname, fullname)
        :return: True to include the user by default, False to exclude
        """
        hps = api.content.find(
            portal_type="held_position",
            userid=user_data["userid"],
        )
        for hp in hps:
            hp_obj = hp.getObject()
            if base_hasattr(hp_obj, "usages") and "signer" in hp_obj.usages:
                return True

        user_obj = api.user.get(userid=user_data["userid"])
        if user_obj:
            for group in api.group.get_groups(user=user_obj):
                if group.getId().endswith("watchers"):
                    return True

        return False

    def get_users_data(self):
        """Get all users data sorted by filter status then userid.

        :return: list of user data dictionaries with 'checked' status
        """
        fn_first = True
        portal = api.portal.get()
        catalog = getToolByName(portal, "portal_catalog")
        acl_users = getToolByName(portal, "acl_users")

        all_users_data = []
        email_registry = {}

        for user_info in acl_users.searchUsers():
            userid = user_info.get("userid")
            if not userid:
                continue

            # Skip duplicates
            if any(u["userid"] == userid for u in all_users_data):
                continue

            user_obj = api.user.get(userid=userid)
            if not user_obj:
                continue

            email = user_obj.getProperty("email", "")
            fullname = user_obj.getProperty("fullname", "")
            lastname = firstname = ""

            # Do we have a person with this userid?
            brains = catalog.searchResults(portal_type="person", userid=userid)
            if brains:
                person = brains[0].getObject()
                lastname = getattr(person, "lastname", "") or ""
                firstname = getattr(person, "firstname", "") or ""

            if not lastname and not firstname:
                start = api.portal.get_registry_record(
                    "imio.dms.mail.browser.settings.IImioDmsMailConfig.omail_fullname_used_form", default=None
                )
                if start is not None:
                    fn_first = start == "firstname"
                firstname, lastname = separate_fullname(user_obj, fn_first=fn_first)

            user_data = {
                "userid": userid,
                "email": email,
                "lastname": lastname,
                "firstname": firstname,
                "fullname": fullname,
            }

            # Determine if user should be checked by default
            user_data["checked"] = self.filter_user(user_data)

            all_users_data.append(user_data)

            if email:
                email_registry.setdefault(email, []).append(userid)

        # Calculate duplicates
        duplicates = {email: userids for email, userids in email_registry.items() if len(userids) > 1}

        # Mark users with duplicate emails
        for user_data in all_users_data:
            user_data["has_duplicate_email"] = user_data["email"] in duplicates

        # Sort: filtered users first (checked=True), then by userid
        all_users_data.sort(key=lambda x: (not x["checked"], x["userid"]))

        return all_users_data, duplicates

    def _get_selected_userids(self):
        """Get list of selected user IDs from request.

        Expects a JSON-formatted list in 'selected_users' parameter.
        Returns an empty list if the input is not valid JSON.
        """
        selected = self.request.get("selected_users", "")

        if not selected:
            return []

        # Parse JSON array
        try:
            return json.loads(selected)
        except (ValueError, TypeError):
            return []

    def _download_csv(self):
        """Generate and download CSV file with selected users."""
        selected_userids = self._get_selected_userids()

        if not selected_userids:
            api.portal.show_message(_("No users selected for CSV download."), request=self.request, type="warning")
            return self.request.RESPONSE.redirect(self.context.absolute_url() + "/@@signing-users-csv")

        all_users_data, __ = self.get_users_data()

        # Filter to only selected users
        selected_users = [u for u in all_users_data if u["userid"] in selected_userids]

        # Generate CSV
        csv_output = StringIO()
        writer = csv.DictWriter(
            csv_output,
            fieldnames=["userid", "email", "lastname", "firstname", "fullname"],
            delimiter=",",
            quoting=csv.QUOTE_MINIMAL,
        )
        writer.writeheader()
        for user_data in selected_users:
            writer.writerow(
                {
                    "userid": safe_encode(user_data["userid"]),
                    "email": safe_encode(user_data["email"]),
                    "lastname": safe_encode(user_data["lastname"]),
                    "firstname": safe_encode(user_data["firstname"]),
                    "fullname": safe_encode(user_data["fullname"]),
                }
            )

        output = csv_output.getvalue()
        response = self.request.RESPONSE
        response.setHeader("Content-Type", "text/csv; charset=utf-8")
        response.setHeader("Content-Disposition", "attachment; filename=signing_users_selected.csv")
        return output

    def _send_emails(self):
        """Send emails to selected users."""
        selected_userids = self._get_selected_userids()

        if not selected_userids:
            api.portal.show_message(_("No users selected for email sending."), request=self.request, type="warning")
            return self.request.RESPONSE.redirect(self.context.absolute_url() + "/@@signing-users-csv")

        email_content = get_esign_registry_signing_users_email_content()
        if not email_content:
            api.portal.show_message(
                _("Email content is not configured in the settings."), request=self.request, type="error"
            )
            return self.request.RESPONSE.redirect(self.context.absolute_url() + "/@@signing-users-csv")

        all_users_data, _duplicates = self.get_users_data()
        selected_users = [u for u in all_users_data if u["userid"] in selected_userids]

        portal_email = api.portal.get().getProperty("email_from_address")
        if not portal_email:
            api.portal.show_message(_("Portal from email is not configured."), request=self.request, type="error")
            return self.request.RESPONSE.redirect(self.context.absolute_url() + "/@@mail-controlpanel")

        success_count = 0
        failed_count = 0
        for user_data in selected_users:
            if not user_data["email"]:
                failed_count += 1
                api.portal.show_message(
                    _(
                        "User ${userid} has no email address configured. Skipping.",
                        mapping={"userid": user_data["userid"]},
                    ),
                    request=self.request,
                    type="warning",
                )
                continue

            personalized_content = self._render_email_content(email_content, user_data)
            # personalized_content = str(email_content).format(**user_data).replace("\n", "<br>\n")

            # Create and send email
            try:
                eml = create_html_email(personalized_content, with_plain=True)
                subject = translate(_(u"You have been invited to Paraphéo"), context=self.request)

                status, error = send_email(
                    eml, subject=subject, mfrom=portal_email, mto=user_data["email"], immediate=False
                )

                if status:
                    success_count += 1
                else:
                    raise Exception(error)
            except Exception as e:
                failed_count += 1
                error = str(e)
                api.portal.show_message(
                    _(
                        "Failed to send email to ${userid}.",
                        mapping={"userid": user_data["userid"]},
                    )
                    + " "
                    + error,
                    request=self.request,
                    type="error",
                )
                continue

        if success_count > 0:
            api.portal.show_message(
                _("Emails sent successfully to ${count} users.", mapping={"count": success_count}),
                request=self.request,
                type="info",
            )

        if failed_count > 0:
            api.portal.show_message(
                _("Failed to send emails to ${count} users.", mapping={"count": failed_count}),
                request=self.request,
                type="warning",
            )

        return self.request.RESPONSE.redirect(self.context.absolute_url() + "/@@signing-users-csv")

    def _render_email_content(self, template, user_data):
        """Render the email content template with user data.

        :param template: The email content template (TAL compliant)
        :param user_data: dict containing user data (userid, email, lastname, firstname, fullname)
        :return: Rendered email content as a string
        """
        pt = PageTemplate()
        pt.pt_source_file = lambda: "none"
        pt.write(template)
        namespace = pt.pt_getContext()
        namespace.update(
            {
                "request": self.request,
                "view": self,
                "context": self.context,
                "user_data": user_data,
                "parapheo_url": get_esign_registry_parapheo_url(),
                "modules": SecureModuleImporter,
            }
        )
        return pt.pt_render(namespace)


class EsignMacros(BrowserView):
    """ """



# -*- coding: utf-8 -*-
from imio.fpaudit import utils as _fpaudit_utils
from plone.app.robotframework.testing import REMOTE_LIBRARY_BUNDLE_FIXTURE
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.testing import z2
from zope.globalrequest import setLocal

import imio.esign  # noqa: F401


logged_actions = []


def _mock_fpalog(log_id, action, extras):
    logged_actions.append(extras)


_fpaudit_utils.fpalog = _mock_fpalog


class ImioEsignLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        # Load any other ZCML that is required for your tests.
        # The z3c.autoinclude feature is disabled in the Plone fixture base
        # layer.
        import plone.app.dexterity

        self.loadZCML(package=plone.app.dexterity)
        import plone.restapi

        self.loadZCML(package=plone.restapi)
        import imio.annex

        self.loadZCML(package=imio.annex)
        self.loadZCML(package=imio.esign)

    def setUpPloneSite(self, portal):
        setLocal("request", portal.REQUEST)
        applyProfile(portal, "imio.annex:default")
        applyProfile(portal, "imio.esign:default")


IMIO_ESIGN_FIXTURE = ImioEsignLayer()


IMIO_ESIGN_INTEGRATION_TESTING = IntegrationTesting(
    bases=(IMIO_ESIGN_FIXTURE,),
    name="ImioEsignLayer:IntegrationTesting",
)


IMIO_ESIGN_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(IMIO_ESIGN_FIXTURE,),
    name="ImioEsignLayer:FunctionalTesting",
)


IMIO_ESIGN_ACCEPTANCE_TESTING = FunctionalTesting(
    bases=(
        IMIO_ESIGN_FIXTURE,
        REMOTE_LIBRARY_BUNDLE_FIXTURE,
        z2.ZSERVER_FIXTURE,
    ),
    name="ImioEsignLayer:AcceptanceTesting",
)



# -*- coding: utf-8 -*-
from datetime import datetime
from imio.esign import logger
from imio.esign.audit import audit
from imio.esign.utils import get_session_annotation
from imio.helpers.ws import verify_auth_token
from plone.restapi.deserializer import json_body
from plone.restapi.services import Service


class ExternalSessionFeedbackPost(Service):
    def reply(self):  # noqa C901
        """Handle the external session feedback.

        Needs json body with:
            * "app_session_id": int 1234560001, app_session_id
            * "code": int some_code, feedback identification code
            * "session_state": microservice session state
            * "value": json dict contaaining sign URL or signer/refused emails
            * "message": "some message", optional message with feedback
        """
        if not self.authorized():
            self.request.response.setStatus(403)
            return {"message": "Unauthorized access"}
        data = json_body(self.request)
        app_session_id = data.get("app_session_id")
        logger.info("External session feedback received: {}".format(data))
        if not app_session_id:
            self.request.response.setStatus(400)
            return {"message": "app_session_id is required"}
        code = int(data.get("code"))
        if not code:
            self.request.response.setStatus(400)
            return {"message": "code is required"}
        value = data.get("value") or {}
        db_state = data.get("session_state")
        try:
            annot = get_session_annotation()
            session_id = int(app_session_id[7:])
            if session_id not in annot["sessions"]:
                self.request.response.setStatus(400)
                return {"message": "Session ID {} not found".format(session_id)}
            session = annot["sessions"][session_id]
            session_update = {"returns": session["returns"]}
            session_update["returns"].append((code, db_state, data.get("value", ""), data.get("message", ""),
                                              datetime.now()))
            if code == 21:
                # sign_session_confirmed
                session_update["state"] = "to_sign"
                if value and "sign_session_url" in value and not session["sign_url"]:
                    session_update["sign_url"] = value["sign_session_url"]
            elif code == 22:
                # one_signer_accepted
                if value and "signed_users" in value:
                    session_update["signers"] = session["signers"]
                    for i, d in enumerate(session["signers"]):
                        if d["status"] in ("signed", "refused"):
                            continue
                        if d["email"] in value["signed_users"]:
                            session_update["signers"][i]["status"] = "signed"
            elif code == 23:
                # upload_success (files returned)
                session_update["state"] = "returned"
            elif code == 52:
                # one_signer_refused
                session_update["state"] = "refused"
                if value and "user" in value:
                    session_update["signers"] = session["signers"]
                    for i, d in enumerate(session["signers"]):
                        if d["email"] == value["user"]:
                            session_update["signers"][i]["status"] = "refused"
                            break
            elif code == 53:
                # upload_failed
                session_update["state"] = "signed"
            elif code in (50, 40, 51, 41):
                # seal_creation_error, seal_creation_not_available, sign_creation_error, sign_creation_not_available
                session_update["state"] = "errored"
            if session_update:
                session.update(session_update)
                session["last_update"] = datetime.now()
            audit("session_feedback", "session={} code={} db_state={}".format(session_id, code, db_state))

        except Exception as e:
            self.request.response.setStatus(500)
            logger.error(str(e))
            return {"message": str(e)}
        return {"message": "Information correctly handled"}
    """ microservice session state
    to_create_session = "to_create_session"
    session_creation_failed = "session_creation_failed"
    to_sign = "to_sign"
    refused = "refused"
    to_upload = "to_upload"
    to_notify_ged_upload = "to_notify_ged_upload"
    completed = "completed"
    """

    def authorized(self):
        """Check if the user is authorized to access this service."""
        auth_header = getattr(self.request, "_auth", None)
        if not auth_header or not auth_header.startswith("Bearer "):
            return False
        token = auth_header[7:]  # len("Bearer ") == 7
        if not token:
            return False
        return verify_auth_token(token, groups=["access_imio-apps-docs"])



# -*- coding: utf-8 -*-
from imio.esign.audit import LOG_ID
from imio.esign.config import get_esign_registry_parapheo_url
from imio.esign.config import get_esign_registry_signing_users_email_content
from imio.esign.config import set_esign_registry_parapheo_url
from imio.esign.config import set_esign_registry_signing_users_email_content
from imio.esign.config import SIGNERS_EMAIL_CONTENT
from plone.registry.interfaces import IRegistry
from Products.CMFPlone.interfaces import INonInstallable
from zope.component import getUtility
from zope.interface import implementer


@implementer(INonInstallable)
class HiddenProfiles(object):
    def getNonInstallableProfiles(self):
        """Hide uninstall profile from site-creation and quickinstaller."""
        return [
            "imio.esign:uninstall",
        ]

    def getNonInstallableProducts(self):
        """Hide the upgrades package from site-creation and quickinstaller."""
        return ["imio.esign.upgrades"]


def post_install(context):
    """Post install script"""
    if not get_esign_registry_parapheo_url():
        set_esign_registry_parapheo_url(u"https://simplycosi-1-test.trustsigneurope.com/login?tenantName=IMIO")
    if not get_esign_registry_signing_users_email_content():
        set_esign_registry_signing_users_email_content(SIGNERS_EMAIL_CONTENT)
    configure_fpaudit()


def configure_fpaudit():
    """Add esign audit log entry to imio.fpaudit registry if not already present."""
    registry = getUtility(IRegistry)
    entries = list(registry.get("imio.fpaudit.settings.log_entries") or [])
    if not any(e.get("log_id") == LOG_ID for e in entries):
        entries.append({u"log_id": LOG_ID, u"audit_log": u"esign.log",
                        u"log_format": u"%(asctime)s - %(message)s"})
        registry["imio.fpaudit.settings.log_entries"] = entries


def uninstall(context):
    """Uninstall script"""
    # Do something at the end of the uninstallation of this package.



# -*- coding: utf-8 -*-
from AccessControl import Unauthorized
from html import escape
from imio.esign import _
from imio.esign.adapters import ISignable
from imio.esign.audit import audit
from imio.esign.utils import add_files_to_session
from imio.esign.utils import get_session_annotation
from imio.esign.utils import get_sessions_for
from imio.esign.utils import persistent_to_native
from imio.esign.utils import remove_context_from_session
from imio.esign.utils import remove_files_from_session
from imio.helpers.content import uuidToObject
from imio.helpers.security import check_zope_admin
from plone import api
from Products.CMFPlone.utils import safe_unicode
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from six import string_types

import pprint
import re


class AddToSessionView(BrowserView):
    """View to add an element to an esign session."""

    def __init__(self, context, request):
        super(AddToSessionView, self).__init__(context, request)

    def _finished(self, failed_msgid="", mapping={}):
        msgid = "Element added to session!"
        msg_type = "info"
        if failed_msgid:
            msgid = failed_msgid
            msg_type = "warning"
        api.portal.show_message(_(msgid, mapping=mapping), request=self.request, type=msg_type)
        self.request.RESPONSE.redirect(self.request['HTTP_REFERER'])

    def index(self):
        files_uids = ISignable(self.context).get_files_uids()
        if not files_uids:
            return self._finished(failed_msgid="Could not get files uids to add to the session!")
        signers = self.get_signers()
        if not signers:
            return self._finished(failed_msgid="Could not get signers to add to the session!")
        # watchers = self.get_watchers()
        session_id, _session = add_files_to_session(
            signers=signers,
            # watchers=watchers,
            files_uids=files_uids,
            title=self.get_session_title(),
            watchers=self.get_watchers(),
            discriminators=self.get_discriminators(),
        )
        # audit("add_to_session", "session={} context={} files={} signers={}".format(
        #     session_id, self.context.UID(), ",".join(files_uids), len(signers)))
        self._finished()

    def get_signers(self):
        """Get the list of held_positions to be used as signer.

        :return: list of signer infos with "held_position", "name",
                 "function", "userid" and "email"
        """
        try:
            signers = ISignable(self.context).get_signers()
        except ValueError as msg:
            signers = []
            api.portal.show_message(
                _(
                    "Problem getting signers: \"${error}\")!",
                    mapping={"error": str(msg)},
                ),
                request=self.request,
                type="warning",
            )
        return signers

    def get_watchers(self):
        """List of watchers email."""
        return ISignable(self.context).get_watchers()

    def get_context_uid(self):
        """ """
        return self.context.UID()

    def get_session_title(self):
        """The title for the session.

        :return: a string with informative session title
        """
        return "Session title"

    def get_discriminators(self):
        """ """
        return ISignable(self.context).get_discriminators()


class RemoveFromSessionView(BrowserView):
    """View to remove an element from an esign session."""

    def __init__(self, context, request):
        super(RemoveFromSessionView, self).__init__(context, request)

    def _finished(self):
        msg = _("Element removed from session!")
        api.portal.show_message(msg, request=self.request)
        self.request.RESPONSE.redirect(self.request['HTTP_REFERER'])

    def index(self):
        uid = self.get_uid_to_remove()
        str_session_ids = ",".join([str(sid) for sid in get_sessions_for(uid).keys()])
        remove_context_from_session(context_uids=[uid])
        audit("remove_context_from_session", "sessions={} context={}".format(str_session_ids, uid))
        self._finished()

    def get_uid_to_remove(self):
        """ """
        return self.context.UID()

    def available(self):
        """Defines if the action is available or not."""
        annot = get_session_annotation()
        return self.context.UID() in annot.get("c_uids", {})


class RemoveItemFromSessionView(BrowserView):
    """View to remove an item from an esign session."""

    def __init__(self, context, request):
        super(RemoveItemFromSessionView, self).__init__(context, request)

    def _finished(self):
        msg = _("Element removed from session!")
        api.portal.show_message(msg, request=self.request)
        self.request.RESPONSE.redirect(self.request['HTTP_REFERER'])

    def index(self):
        uid = self.context.UID()
        session_id = get_session_annotation().get("uids", {}).get(uid)
        remove_files_from_session(files_uids=[uid])
        audit("remove_item_from_session", "session={} file={}".format(session_id, uid))
        self._finished()

    def available(self):
        """Defines if the action is available or not."""
        annot = get_session_annotation()
        return self.context.UID() in annot.get("uids", {})


class SessionAnnotationInfoView(BrowserView):
    """Admin-only view displaying imio.esign session annotations for a specific context item."""

    index = ViewPageTemplateFile("templates/session_annotation_info.pt")

    def __call__(self):
        if not check_zope_admin():
            raise Unauthorized
        return self.index()

    def _uid_to_link(self, uid):
        """Return an HTML link for an object UID, or the UID if not found."""
        obj = uuidToObject(uid, unrestricted=True)
        if obj is None:
            return u"<span title='not found'>{}</span>".format(safe_unicode(uid))
        url = escape(obj.absolute_url() + "/view", quote=True)
        path = escape(u"/".join(obj.getPhysicalPath()))
        title = escape(safe_unicode(getattr(obj, "title", "") or path))
        return u"<a href='{}' title='{}'>{}</a>".format(url, path, title)

    def _render_value(self, value, indent=u""):
        """Render a value, replacing UIDs with links where possible."""
        inner = indent + u"  "
        if isinstance(value, dict):
            if not value:
                return u"{}"
            lines = [u"{"]
            for k, v in sorted(value.items()):
                key = escape(safe_unicode(pprint.pformat(k)))
                lines.append(u"{}{}: {},".format(inner, key, self._render_value(v, inner)))
            lines.append(u"{}}}".format(indent))
            return u"\n".join(lines)
        elif isinstance(value, (list, tuple)):
            if not value:
                return u"[]"
            lines = [u"["]
            for item in value:
                lines.append(u"{}{},".format(inner, self._render_value(item, inner)))
            lines.append(u"{}]".format(indent))
            return u"\n".join(lines)
        elif isinstance(value, string_types) and re.match(r"^[0-9a-f]{32}$", value):
            # Looks like a UUID
            return self._uid_to_link(value)
        else:
            return escape(safe_unicode(pprint.pformat(value)))

    @property
    def esign_sessions(self):
        """
        Return list of (session_id, session_data) for all sessions.
        Filter sessions using request params "session_id" and "context_uid" if provided.
        Returns all sessions if no filter params provided.
        """
        annot = get_session_annotation()
        request_session_id = self.request.form.get("session_id")
        try:
            request_session_id = int(request_session_id)
        except (ValueError, TypeError):
            request_session_id = None

        c_uid = self.request.form.get("context_uid")
        result = []
        for session_id in annot['sessions']:
            if request_session_id is not None and request_session_id != session_id:
                continue
            session = annot['sessions'][session_id]
            # If any file in this session is in this context
            if c_uid and not any(f['context_uid'] == c_uid for f in session['files']):
                continue
            result.append((session_id, persistent_to_native(session)))
        return sorted(result)

    def esign_session_html(self, session_data):
        """Renders esign session annot in HTML"""
        return self._render_value(session_data)

    def available(self):
        """Defines if the action is available or not."""
        return check_zope_admin()



# -*- coding: utf-8 -*-

from imio.esign import _
from imio.helpers.emailer import validate_email_addresses
from plone.app.registry.browser.controlpanel import ControlPanelFormWrapper
from plone.app.registry.browser.controlpanel import RegistryEditForm
from plone.app.z3cform.wysiwyg import WysiwygFieldWidget
from plone.autoform.directives import widget
from plone.z3cform import layout
from zope import schema
from zope.interface import Interface
from zope.interface import Invalid


def validate_vat_number(va_nb):
    """Validate the VAT number format. It should start with BE followed by 10 digits,
    with the last 2 digits being a control checksum of the first 8 digits."""
    if not va_nb:
        return True

    # Check format: BE followed by 10 digits
    if not va_nb.startswith('BE'):
        raise Invalid(_("VAT number must start with 'BE'"))

    if len(va_nb) != 12:
        raise Invalid(_("VAT number must be 12 characters (BE + 10 digits)"))

    digits = va_nb[2:]
    if not digits.isdigit():
        raise Invalid(_("VAT number must contain 10 digits after 'BE'"))

    # Validate checksum: last 2 digits should be 97 - (first 8 digits modulo 97)
    first_eight = int(digits[:8])
    control = int(digits[8:10])
    expected_control = 97 - (first_eight % 97)

    if control != expected_control:
        raise Invalid(_("Invalid VAT number: checksum verification failed"))

    return True


class IImioEsignSettings(Interface):

    enabled = schema.Bool(
        title=_("Enabled?"),
        description=_("Is the eSign service enabled?"),
        default=True,
    )

    vat_number = schema.TextLine(
        title=_("VAT number"),
        description=_("VAT number used for esign billing (BE0123456789)."),
        constraint=validate_vat_number,
        required=True,
    )

    file_url = schema.URI(
        title=_("File URL download domain"),
        description=_("URL domain where the file can be downloaded."),
        required=False,
        default="https://documents.imio-egov.be/esign",
    )

    seal_code = schema.TextLine(
        title=_("Seal code"),
        description=_("Seal code given by eidas provider."),
        required=False,
    )

    seal_email = schema.TextLine(
        title=_("Seal email"),
        description=_("Email of the eidas provider account containing the seal image."),
        required=False,
    )

    sign_code = schema.TextLine(
        title=_("Sign code"),
        description=_("Sign code used to specify sign method. Keep empty to use default method."),
        required=False,
    )

    parapheo_url = schema.TextLine(
        title=_("Parapheo url"),
        description=_("Used in signers email template."),
        required=False,
    )

    widget("signing_users_email_content", WysiwygFieldWidget)
    signing_users_email_content = schema.Text(
        title=_("Email content model for signing users"),
        description=_(
            "Email content sent to users when inviting them to Parapheo. "
            "TAL compliant with variables: view, context, user_data, parapheo_url, request and modules."
        ),
        required=False,
    )

    max_session_size = schema.Int(
        title=_("Max session size (MB)"),
        description=_("Maximum size of the session in megabytes. If the total size of files to be signed exceeds this "
                      "limit, a new session will be created."),
        default=100,
        min=1,
        required=True,
    )

    external_watchers = schema.TextLine(
        title=_("External watchers emails"),
        description=_("Multiple values must be separated by a comma."),
        constraint=validate_email_addresses,
        required=False,
    )


class ImioEsignSettings(RegistryEditForm):
    schema = IImioEsignSettings
    schema_prefix = "imio.esign"
    label = _("Imio Esign Settings")


ImioEsignSettingsView = layout.wrap_form(
    ImioEsignSettings, ControlPanelFormWrapper
)



# -*- coding: utf-8 -*-
"""utils tests for this package."""
from collections import OrderedDict
from collective.iconifiedcategory.utils import calculate_category_id
from datetime import date
from datetime import timedelta
from imio.esign.config import get_esign_registry_max_session_size
from imio.esign.config import set_esign_registry_external_watchers
from imio.esign.config import set_esign_registry_max_session_size
from imio.esign.testing import IMIO_ESIGN_INTEGRATION_TESTING
from imio.esign.utils import add_files_to_session
from imio.esign.utils import create_external_session
from imio.esign.utils import get_file_download_url
from imio.esign.utils import get_file_info
from imio.esign.utils import get_filesize
from imio.esign.utils import get_max_download_date
from imio.esign.utils import get_session_annotation
from imio.esign.utils import get_session_info
from imio.esign.utils import get_sessions_for
from imio.esign.utils import get_suid_from_uuid
from imio.esign.utils import remove_context_from_session
from imio.esign.utils import remove_files_from_session
from imio.esign.utils import remove_session
from imio.helpers.content import uuidToObject
from imio.pyutils.utils import shortuid_decode_id
from mock import Mock
from mock import patch
from plone import api
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.namedfile.file import NamedBlobFile
from plone.namedfile.file import NamedBlobImage
from zope.annotation import IAnnotations
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent

import collective.iconifiedcategory
import json
import os
import unittest


class TestUtils(unittest.TestCase):

    layer = IMIO_ESIGN_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        # add some users
        api.user.create(email="user1@sign.com", username="user1", password="password1")
        api.user.create(email="user2@sign.com", username="user2", password="password2")
        # add content category configuration
        at_folder = api.content.create(
            container=self.portal,
            id="annexes_types",
            title="Annexes Types",
            type="ContentCategoryConfiguration",
            exclude_from_nav=True,
        )
        category_group = api.content.create(
            type="ContentCategoryGroup",
            title="Annexes",
            container=at_folder,
            id="annexes",
        )
        icon_path = os.path.join(os.path.dirname(collective.iconifiedcategory.__file__), "tests", "icône1.png")
        with open(icon_path, "rb") as fl:
            api.content.create(
                type="ContentCategory",
                title="To sign",
                container=category_group,
                icon=NamedBlobImage(fl.read(), filename=u"icône1.png"),
                id="to_sign",
                predefined_title="To be signed",
                # confidential=True,
                # to_print=True,
                to_sign=True,
                # signed=True,
                # publishable=True,
                # only_pdf=True,
                show_preview=False,
            )
        # add annexes
        self.folders = []
        for f in range(2):
            folder = api.content.create(
                container=self.portal,
                id="folder{}".format(f),
                title="Folder {}".format(f),
                type="Folder",
            )
            self.folders.append(folder)
        tests_dir = os.path.dirname(__file__)
        pdf_files = ["annex1.pdf", "annex2.pdf"]
        self.uids = []
        for i in range(12):
            pdf_file = pdf_files[i % len(pdf_files)]
            container = self.folders[i % len(self.folders)]
            with open(os.path.join(tests_dir, pdf_file), "rb") as f:
                annex = api.content.create(
                    container=container,
                    type="annex",
                    id="annex{}".format(i),
                    title="Annex {}".format(i),
                    content_category=calculate_category_id(self.portal["annexes_types"]["annexes"]["to_sign"]),
                    scan_id="0123456000000{:02d}".format(i),
                    file=NamedBlobFile(data=f.read(), filename=u"annex{}.pdf".format(i), contentType="application/pdf"),
                )
                self.uids.append(annex.UID())

    def test_add_remove_files_to_session(self):
        root_annot = IAnnotations(self.portal)
        self.assertNotIn("imio.esign", root_annot)
        signers = [
            ("user1", "user1@sign.com", "User 1", "Position 1"),
            ("user2", "user2@sign.com", "User 2", "Position 2"),
        ]

        # add files, no session_id, no discriminator
        sid, session = add_files_to_session(signers, (self.uids[0],), title="my title", watchers=("stalker@sign.com",))
        self.assertEqual(sid, 0)
        annot = root_annot["imio.esign"]
        self.assertEqual(annot["numbering"], 1)
        self.assertEqual(len(annot["sessions"]), 1)
        self.assertEqual(len(annot["uids"]), 1)
        self.assertIn(self.uids[0], annot["uids"])
        self.assertEqual(len(annot["c_uids"][self.folders[0].UID()]), 1)
        self.assertDictEqual(dict(annot["c_uids"]), {self.folders[0].UID(): [self.uids[0]]})
        self.assertEqual(session["title"], "my title")
        self.assertEqual(session["state"], "draft")
        self.assertEqual(session["seal"], None)
        self.assertEqual(session["acroform"], True)
        self.assertIsNone(session["sign_url"])
        self.assertEqual(session["client_id"], "0123456")
        self.assertEqual(len(session["watchers"]), 1)
        self.assertEqual(len(session["files"]), 1)
        self.assertListEqual(
            list(session["files"]),
            [
                {
                    "context_uid": self.folders[0].UID(),
                    "scan_id": "012345600000000",
                    "title": "Annex 0",
                    "uid": self.uids[0],
                    "status": "",
                    "filename": u"annex0.pdf",
                }
            ],
        )
        self.assertEqual(len(session["signers"]), 2)
        self.assertEqual(session["size"], 6968)  # annex1.pdf

        # add files, no session_id => same session reused
        signers[1] = ("user2", "user2@sign.com", "User 2", "Position 2b")  # changed position => not discriminant
        # rename uids[1] to same name as uids[0] to test duplicate filename handling
        annex1_obj = uuidToObject(uuid=self.uids[1], unrestricted=True)
        annex1_obj.file.filename = u"annex0.pdf"
        sid, session = add_files_to_session(signers, (self.uids[1],))
        self.assertEqual(sid, 0)
        self.assertEqual(annot["numbering"], 1)
        self.assertEqual(len(annot["sessions"]), 1)
        self.assertEqual(len(annot["uids"]), 2)
        self.assertIn(self.uids[1], annot["uids"])
        self.assertEqual(len(annot["c_uids"]), 2)
        self.assertIn(self.folders[1].UID(), annot["c_uids"])
        self.assertEqual(len(session["files"]), 2)
        # verify duplicate filename was renamed
        self.assertEqual(session["files"][1]["filename"], u"annex0-1.pdf")
        self.assertEqual(len(annot["c_uids"][self.folders[1].UID()]), 1)
        self.assertEqual(session["size"], 6968 + 7014)  # annex1 + annex2

        # add files, no session_id, new discriminations => new session
        sid, session = add_files_to_session(signers, (self.uids[2],), discriminators=("council1",))
        self.assertEqual(sid, 1)
        self.assertEqual(annot["numbering"], 2)
        self.assertEqual(len(annot["sessions"]), 2)
        self.assertEqual(len(annot["uids"]), 3)
        self.assertIn(self.uids[2], annot["uids"])
        self.assertEqual(len(session["files"]), 1)
        self.assertEqual(len(session["watchers"]), 0)

        # add files, no session_id, same discriminations => same session
        sid, session = add_files_to_session(signers, (self.uids[3],), discriminators=("council1",))
        self.assertEqual(sid, 1)
        self.assertEqual(annot["numbering"], 2)
        self.assertEqual(len(annot["sessions"]), 2)
        self.assertEqual(len(annot["uids"]), 4)
        self.assertIn(self.uids[3], annot["uids"])
        self.assertEqual(len(session["files"]), 2)

        # add files, no session_id, other discriminations => other session
        sid, session = add_files_to_session(signers, (self.uids[4],), discriminators=("council2",))
        self.assertEqual(sid, 2)

        # add files, session_id, other discriminations => reused session
        sid, session = add_files_to_session(signers, (self.uids[5],), session_id=0, discriminators=("council3",))
        self.assertEqual(sid, 0)

        # add files, session_id unfound, other discriminations => new session
        sid, session = add_files_to_session(signers, (self.uids[6],), session_id=999, discriminators=("council3",))
        self.assertEqual(sid, 3)

        # add files, no session_id, different signers => new session
        sid, session = add_files_to_session([signers[0]], (self.uids[7],))
        self.assertEqual(sid, 4)

        # add files, no session_id, different seal => new session
        sid, session = add_files_to_session(signers, (self.uids[8],), seal="seal1")
        self.assertEqual(sid, 5)

        # add files, no session_id, same seal, different acroform => new session
        sid, session = add_files_to_session(signers, (self.uids[9],), acroform=False)
        self.assertEqual(sid, 6)
        self.assertEqual(len(annot["uids"]), 10)
        self.assertEqual(len(annot["c_uids"]), 2)
        self.assertEqual(len(annot["c_uids"][self.folders[0].UID()]), 5)
        self.assertEqual(len(annot["c_uids"][self.folders[1].UID()]), 5)
        self.assertEqual(len(annot["sessions"]), 7)

        # add files, no session_id, no signers => new session
        sid, session = add_files_to_session([], (self.uids[10],), seal="seal2")
        self.assertEqual(sid, 7)
        self.assertEqual(len(annot["sessions"]), 8)
        self.assertEqual(session["signers"], [])
        self.assertEqual(session["seal"], "seal2")

        # add files, no session_id, same seal, different states => new session
        session["state"] = "sent"
        sid, session = add_files_to_session([], (self.uids[11],), seal="seal2")
        self.assertEqual(sid, 8)
        self.assertEqual(len(annot["sessions"]), 9)

        # now we can start to remove
        remove_files_from_session((self.uids[0], self.uids[1]))  # 2 of 3 session files
        self.assertEqual(len(annot["uids"]), 10)
        self.assertEqual(len(annot["sessions"][0]["files"]), 1)
        self.assertEqual(annot["sessions"][0]["size"], 7014)  # only uid[5] (annex2) remains
        self.assertEqual(len(annot["c_uids"][self.folders[0].UID()]), 5)
        self.assertEqual(len(annot["c_uids"][self.folders[1].UID()]), 5)
        remove_files_from_session((self.uids[5],))  # no more session files, session removed
        self.assertEqual(len(annot["uids"]), 9)
        self.assertEqual(len(annot["sessions"]), 8)
        self.assertNotIn(0, annot["sessions"])
        remove_files_from_session((self.uids[2], self.uids[3]))  # all session files, session removed
        self.assertEqual(len(annot["uids"]), 7)
        self.assertEqual(len(annot["sessions"]), 7)
        self.assertNotIn(1, annot["sessions"])
        remove_files_from_session((self.uids[4],))
        remove_files_from_session((self.uids[6],))
        remove_files_from_session((self.uids[7],))
        remove_files_from_session((self.uids[8],))
        remove_files_from_session((self.uids[9],))
        remove_files_from_session((self.uids[10],))
        remove_files_from_session((self.uids[11],))
        self.assertEqual(len(annot["uids"]), 0)
        self.assertEqual(len(annot["c_uids"]), 0)
        self.assertEqual(len(annot["sessions"]), 0)

    def test_create_external_session(self):
        """Test create_external_session with mocked requests.post and get_auth_token."""
        from mock import MagicMock
        from mock import patch

        # Case 1: session not found => returns sentinel string, no HTTP call made
        with patch('imio.esign.utils.get_auth_token', return_value='test_token'), \
             patch('imio.esign.utils.requests') as mock_requests:
            result = create_external_session(999)
        self.assertEqual(result, "_session_not_found_")
        mock_requests.post.assert_not_called()

        # Setup: create a session with two signers and a watcher
        signers = [
            ("user1", "user1@sign.com", "User 1", "Position 1"),
            ("user2", "user2@sign.com", "User 2", "Position 2"),
        ]
        sid, session = add_files_to_session(
            signers, (self.uids[0], self.uids[2],), title=u"Test session", watchers=("watcher@sign.com",)
        )
        self.assertEqual(sid, 0)
        self.assertEqual(session["state"], "draft")

        # Case 2: HTTP 200 => state becomes "sent", request payload is verified
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.text = '{"message": "Request received"}'
        with patch('imio.esign.utils.get_auth_token', return_value='test_token'), \
             patch('imio.esign.utils.requests') as mock_requests:
            mock_requests.post.return_value = mock_response
            result = create_external_session(sid)
        self.assertIs(result, mock_response)
        self.assertEqual(session["state"], "sent")
        mock_requests.post.assert_called_once()
        call_args = mock_requests.post.call_args
        posted_url = call_args[0][0]
        self.assertIn("imio/esign/v1/luxtrust/sessions", posted_url)
        self.assertEqual(call_args[1]["headers"]["Authorization"], "Bearer test_token")
        data = json.loads(call_args[1]["data"]["data"])
        self.assertIn("commonData", data)
        self.assertIn("signData", data)
        self.assertNotIn("sealData", data)
        self.assertEqual(data["commonData"]["sessionName"], u"Test session")
        self.assertEqual(data["commonData"]["imioAppSessionId"], session["sign_id"])
        self.assertEqual(data["signData"]["users"], ["user1@sign.com", "user2@sign.com"])
        self.assertEqual(data["signData"]["watchers"], ["watcher@sign.com"])
        files_param = call_args[1]["files"]
        self.assertEqual(len(files_param), 2)
        self.assertEqual(files_param[0][0], "files")
        self.assertEqual(files_param[0][1][0], u"annex0.pdf")
        self.assertEqual(files_param[1][0], "files")
        self.assertEqual(files_param[1][1][0], u"annex2.pdf")

        # Case 3: HTTP non-200 => state unchanged (stays "draft")
        sid2, session2 = add_files_to_session(signers, (self.uids[2],), title=u"Test session 2")
        mock_response_fail = MagicMock()
        mock_response_fail.status_code = 400
        mock_response_fail.text = '{"error": "Bad request"}'
        with patch('imio.esign.utils.get_auth_token', return_value='test_token'), \
             patch('imio.esign.utils.requests') as mock_requests:
            mock_requests.post.return_value = mock_response_fail
            result = create_external_session(sid2)
        self.assertIs(result, mock_response_fail)
        self.assertEqual(session2["state"], "draft")

        # Case 4: seal session without seal_code configured => returns "_no_seal_code_", no HTTP call
        api.portal.set_registry_record("imio.esign.seal_email", u"seal@example.com")
        annex0_obj = uuidToObject(uuid=self.uids[0], unrestricted=True)
        annex0_obj.file.filename = u"annex1.pdf"
        sid3, session3 = add_files_to_session(signers, (self.uids[4], self.uids[1], self.uids[0]), seal="PADES_SEAL")
        with patch('imio.esign.utils.get_auth_token', return_value='test_token'), \
             patch('imio.esign.utils.requests') as mock_requests:
            result = create_external_session(sid3)
        self.assertEqual(result, "_no_seal_code_")
        mock_requests.post.assert_not_called()

        # Case 5: seal session with seal_code, custom URL => sealData in payload, correct URL used
        api.portal.set_registry_record("imio.esign.seal_code", u"PADES_SEAL")
        api.portal.set_registry_record("imio.esign.seal_email", u"seal@example.com")
        mock_response_seal = MagicMock()
        mock_response_seal.status_code = 200
        mock_response_seal.text = '{"message": "OK"}'
        with patch('imio.esign.utils.get_auth_token', return_value='test_token'), \
             patch('imio.esign.utils.requests') as mock_requests:
            mock_requests.post.return_value = mock_response_seal
            result = create_external_session(sid3, esign_root_url="https://custom.example.com")
        self.assertIs(result, mock_response_seal)
        self.assertEqual(session3["state"], "sent")
        call_args = mock_requests.post.call_args
        self.assertEqual(call_args[0][0], "https://custom.example.com/imio/esign/v1/luxtrust/sessions")
        data = json.loads(call_args[1]["data"]["data"])
        self.assertIn("sealData", data)
        self.assertEqual(data["sealData"]["sealCode"], "PADES_SEAL")
        self.assertEqual(data["sealData"]["users"], ["seal@example.com"])
        self.assertTrue(data["sealData"]["acroform"])
        files_param = call_args[1]["files"]
        self.assertEqual(len(files_param), 3)
        self.assertEqual(files_param[0][0], "files")
        self.assertEqual(files_param[0][1][0], u"annex4.pdf")
        self.assertEqual(files_param[1][0], "files")
        self.assertEqual(files_param[1][1][0], u"annex1.pdf")
        self.assertEqual(files_param[2][1][0], u"annex1-1.pdf")

        # Case 6: session without files to send => returns "_no_files_", no HTTP call
        for i in range(len(session3["files"])):
            session3["files"][i]["uid"] = "nonexistent_uid_{}".format(i)
        with patch('imio.esign.utils.get_auth_token', return_value='test_token'), \
             patch('imio.esign.utils.requests') as mock_requests:
            result = create_external_session(sid3)
        self.assertEqual(result, "_no_files_")
        mock_requests.post.assert_not_called()

        # case 7: bad session number to send => returns "_session_not_found_", no HTTP call
        with patch('imio.esign.utils.get_auth_token', return_value='test_token'), \
             patch('imio.esign.utils.requests') as mock_requests:
            result = create_external_session(99)
        self.assertEqual(result, "_session_not_found_")
        mock_requests.post.assert_not_called()

    def test_get_filesize(self):
        """Test get_filesize returns the correct file size."""
        # even index => annex1.pdf (6968 bytes)
        self.assertEqual(get_filesize(self.uids[0]), 6968)
        # odd index => annex2.pdf (7014 bytes)
        self.assertEqual(get_filesize(self.uids[1]), 7014)
        # invalid UID returns 0
        self.assertEqual(get_filesize("nonexistent_uid"), 0)
        # check size get robustness
        annex = uuidToObject(uuid=self.uids[0], unrestricted=True)
        self.assertEqual(annex.content_category, "plone-annexes_types_-_annexes_-_to_sign")
        folder = annex.__parent__
        self.assertEqual(folder.categorized_elements[self.uids[0]]["filesize"], 6968)
        folder.categorized_elements[self.uids[0]]["filesize"] = 1111
        self.assertEqual(get_filesize(self.uids[0]), 1111)
        del folder.categorized_elements[self.uids[0]]["filesize"]
        self.assertEqual(get_filesize(self.uids[0]), 6968)
        del folder.categorized_elements[self.uids[0]]
        self.assertEqual(get_filesize(self.uids[0]), 6968)
        delattr(folder, "categorized_elements")
        self.assertEqual(get_filesize(self.uids[0]), 6968)

    def test_session_size_discrimination(self):
        """Test that sessions are split when they would exceed max_session_size."""
        signers = [
            ("user1", "user1@sign.com", "User 1", "Position 1"),
        ]

        # add one file to create a session (annex1.pdf = 6968 bytes)
        sid, session = add_files_to_session(signers, (self.uids[0],))
        self.assertEqual(sid, 0)
        self.assertEqual(session["size"], 6968)

        # set session size just under the 1 MB limit so adding another file exceeds it
        session["size"] = 1 * 1024**2 - 1  # 1 MB - 1 byte
        previous_max = get_esign_registry_max_session_size()
        self.addCleanup(set_esign_registry_max_session_size, previous_max)
        set_esign_registry_max_session_size(1)

        # adding a file (~7KB) would exceed 1 MB => new session created
        sid2, session2 = add_files_to_session(signers, (self.uids[1],))
        self.assertEqual(sid2, 1)
        self.assertIsNot(session2, session)
        self.assertEqual(session2["size"], 7014)

        # with enough room, files go to the same session
        sid3, session3 = add_files_to_session(signers, (self.uids[2],))
        self.assertEqual(sid3, 1)
        self.assertIs(session3, session2)
        self.assertEqual(session3["size"], 7014 + 6968)

    def test_add_files_with_duplicate_filenames(self):
        """Test that files with duplicate filenames are renamed with suffix."""
        annot = get_session_annotation()
        self.assertEqual(len(annot["sessions"]), 0)

        signers = [
            ("user1", "user1@sign.com", "User 1", "Position 1"),
        ]

        for i in range(3):
            annex = api.content.get(UID=self.uids[i])
            annex.file.filename = u"same_filename.pdf"
            # annex.reindexObject()

        sid, session = add_files_to_session(signers, (self.uids[0],))
        self.assertEqual(len(session["files"]), 1)
        self.assertEqual(session["files"][0]["filename"], "same_filename.pdf")
        sid, session = add_files_to_session(signers, (self.uids[1],), session_id=sid)
        self.assertEqual(len(session["files"]), 2)
        self.assertIn("same_filename-1.pdf", [f["filename"] for f in session["files"]])

        sid, session = add_files_to_session(signers, (self.uids[2],), session_id=sid)
        self.assertEqual(len(session["files"]), 3)
        filenames = [f["filename"] for f in session["files"]]
        self.assertIn("same_filename.pdf", filenames)
        self.assertIn("same_filename-1.pdf", filenames)
        self.assertIn("same_filename-2.pdf", filenames)

    def test_add_files_already_exist_is_updated(self):
        """When adding a file to the same session, it is updated."""
        annot = get_session_annotation()
        self.assertEqual(len(annot["sessions"]), 0)

        signers = [
            ("user1", "user1@sign.com", "User 1", "Position 1"),
        ]

        annex0_uid = self.uids[0]
        annex0 = api.content.get(UID=annex0_uid)

        sid, session = add_files_to_session(signers, (annex0_uid,))
        self.assertEqual(sid, 0)
        self.assertEqual(len(session["files"]), 1)
        self.assertEqual(session["files"][0]["filename"], "annex0.pdf")
        self.assertEqual(session["files"][0]["title"], "Annex 0")
        self.assertEqual(session["size"], 6968)
        # edit annex and add again, still one annex in session and data are updated
        annex0.file.filename = u"new_annex0.pdf"
        annex0.setTitle('New Annex 0')
        sid, session = add_files_to_session(signers, (annex0_uid,))
        # same session_id
        self.assertEqual(sid, 0)
        self.assertEqual(len(session["files"]), 1)
        self.assertEqual(session["files"][0]["filename"], "new_annex0.pdf")
        self.assertEqual(session["files"][0]["title"], "New Annex 0")
        self.assertEqual(session["size"], 6968)
        # add again exact same file
        sid, session = add_files_to_session(signers, (annex0_uid,))
        self.assertEqual(sid, 0)
        self.assertEqual(len(session["files"]), 1)
        self.assertEqual(session["files"][0]["filename"], "new_annex0.pdf")
        self.assertEqual(session["files"][0]["title"], "New Annex 0")
        self.assertEqual(session["size"], 6968)
        # add second file 2 times
        annex1_uid = self.uids[1]
        annex1 = api.content.get(UID=annex1_uid)
        sid, session = add_files_to_session(signers, (annex1_uid,))
        self.assertEqual(sid, 0)
        self.assertEqual(len(session["files"]), 2)
        self.assertEqual(session["files"][1]["filename"], "annex1.pdf")
        self.assertEqual(session["files"][1]["title"], "Annex 1")
        self.assertEqual(session["size"], 13982)
        # edit and add again
        annex1.setTitle('New Annex 1')
        # edit file, filename and content so size changed
        with open(os.path.join(os.path.dirname(__file__), "annex1.pdf"), "rb") as f:
            annex1.file = NamedBlobFile(data=f.read(), filename=u"new_annex1.pdf", contentType="application/pdf")
        notify(ObjectModifiedEvent(annex1))
        # data were already updated
        self.assertEqual(len(session["files"]), 2)
        self.assertEqual(session["files"][1]["filename"], "new_annex1.pdf")
        self.assertEqual(session["files"][1]["title"], "New Annex 1")
        self.assertEqual(session["files"][1]["scan_id"], "012345600000001")
        self.assertEqual(session["size"], 13936)
        # change file but add a filename already used, change scan_id as well
        with open(os.path.join(os.path.dirname(__file__), "annex1.pdf"), "rb") as f:
            annex1.file = NamedBlobFile(data=f.read(), filename=u"new_annex0.pdf", contentType="application/pdf")
        annex1.scan_id = "012345600000002"
        notify(ObjectModifiedEvent(annex1))
        self.assertEqual(session["files"][1]["filename"], "new_annex0-1.pdf")
        self.assertEqual(session["files"][1]["scan_id"], "012345600000002")
        # edit annex out of any session
        annex2_uid = self.uids[2]
        annex2 = api.content.get(UID=annex2_uid)
        notify(ObjectModifiedEvent(annex2))
        # just to check, remove annex0
        remove_files_from_session((annex0_uid,))
        self.assertEqual(session["size"], 6968)

    def test_remove_context_from_session(self):
        """Test removing a context from a session."""
        annot = get_session_annotation()
        self.assertEqual(len(annot["sessions"]), 0)
        signers = [
            ("user1", "user1@sign.com", "User 1", "Position 1"),
            ("user2", "user2@sign.com", "User 2", "Position 2"),
        ]
        sid, session = add_files_to_session(signers, (self.uids[0], self.uids[1], self.uids[2], self.uids[3]))
        self.assertEqual(len(annot["uids"]), 4)
        self.assertEqual(len(annot["c_uids"]), 2)
        self.assertEqual(len(annot["sessions"]), 1)
        remove_context_from_session((self.folders[0].UID(),))
        self.assertEqual(len(annot["uids"]), 2)
        self.assertEqual(len(annot["c_uids"]), 1)
        self.assertEqual(len(annot["sessions"]), 1)
        remove_context_from_session((self.folders[1].UID(),))
        self.assertEqual(len(annot["uids"]), 0)
        self.assertEqual(len(annot["c_uids"]), 0)
        self.assertEqual(len(annot["sessions"]), 0)

    def test_remove_session(self):
        """Test removing a session."""
        annot = get_session_annotation()
        self.assertEqual(len(annot["sessions"]), 0)
        signers = [
            ("user1", "user1@sign.com", "User 1", "Position 1"),
            ("user2", "user2@sign.com", "User 2", "Position 2"),
        ]
        sid, session = add_files_to_session(signers, (self.uids[0], self.uids[1]))
        self.assertEqual(sid, 0)
        sid, session = add_files_to_session(signers, (self.uids[2], self.uids[3]), seal="seal1")
        self.assertEqual(sid, 1)
        remove_session(0)  # remove first session
        self.assertEqual(len(annot["uids"]), 2)
        self.assertEqual(len(annot["c_uids"]), 2)
        self.assertEqual(len(annot["sessions"]), 1)

    def test_get_session_info(self):
        """Test getting info about a given session id."""
        annot = get_session_annotation()
        self.assertEqual(len(annot["sessions"]), 0)
        self.assertEqual(get_session_info(0), {})
        self.assertEqual(get_session_info(1), {})
        self.assertEqual(get_session_info(2), {})
        signers = [
            ("user1", "user1@sign.com", "User 1", "Position 1"),
            ("user2", "user2@sign.com", "User 2", "Position 2"),
        ]
        sid, session = add_files_to_session(signers, (self.uids[0], self.uids[1]))
        self.assertEqual(get_session_info(sid), session)

    def test_get_sessions_for(self):
        """Test getting sessions for a given context_uid."""
        # no session
        context_uid = self.folders[0].UID()
        self.assertEqual(get_sessions_for(context_uid), OrderedDict())
        # one session
        signers = [("user1", "user1@sign.com", "User 1", "Position 1")]
        sid, session = add_files_to_session(signers, (self.uids[0],))
        self.assertEqual(get_sessions_for(context_uid).keys(), [0])
        # two sessions
        signers = [("user2", "user2@sign.com", "User 2", "Position 2")]
        sid, session = add_files_to_session(signers, (self.uids[2],))
        self.assertEqual(get_sessions_for(context_uid).keys(), [0, 1])
        # readonly=True
        sessions = get_sessions_for(context_uid)
        self.assertEqual(get_session_info(0)['watchers'], [])
        sessions[0]['watchers'] = ["watcher@sign.com"]
        self.assertEqual(get_session_info(0)['watchers'], [])
        # readonly=False
        sessions = get_sessions_for(context_uid, readonly=False)
        self.assertEqual(get_session_info(0)['watchers'], [])
        sessions[0]['watchers'] = ["watcher@sign.com"]
        self.assertEqual(get_session_info(0)['watchers'], ["watcher@sign.com"])

    def test_get_file_info(self):
        """Test getting infos for a given file."""
        annex0_uid = self.uids[0]
        annex1_uid = self.uids[1]
        # no session
        self.assertIsNone(get_file_info(0, annex0_uid))
        # create session
        signers = [("user1", "user1@sign.com", "User 1", "Position 1")]
        sid, session = add_files_to_session(signers, (annex0_uid,))
        self.assertEqual(get_file_info(0, annex0_uid)['uid'], annex0_uid)
        self.assertIsNone(get_file_info(0, annex1_uid))
        # readonly=True by default
        file_info = get_file_info(0, annex0_uid)
        file_info['title'] = u'New title annex 0'
        # not changed in the annotation
        self.assertEqual(
            get_session_annotation()['sessions'][0]['files'][0]['title'], u'Annex 0')
        # readonly=False
        file_info = get_file_info(0, annex0_uid, readonly=False)
        file_info['title'] = u'New title annex 0'
        # changed in the annotation
        self.assertEqual(
            get_session_annotation()['sessions'][0]['files'][0]['title'], u'New title annex 0')

    def test_get_file_download_url(self):
        """Test generating file download URL from UID."""
        uid = "f40682caafc045b4b81973bd82ea9ab6"

        api.portal.set_registry_record("imio.esign.file_url", "https://downloads.files.com")

        result = get_file_download_url(uid)
        self.assertEqual(result, ("https://downloads.files.com/Rzgwy-9BVG9-viEts-5GBkn-Rm",
                                  "Rzgwy-9BVG9-viEts-5GBkn-Rm"))

        custom_url = "https://custom.domain.org/"
        result = get_file_download_url(uid, root_url=custom_url)
        self.assertEqual(result[0], "https://custom.domain.org/Rzgwy-9BVG9-viEts-5GBkn-Rm")

        # Test with another UID to verify encoding works
        uid2 = self.uids[0]
        result2 = get_file_download_url(uid2)
        self.assertTrue(result2[0].startswith("https://downloads.files.com/"))
        self.assertEqual(shortuid_decode_id(result2[1], separator="-"), uid2)  # correctly decoded

        # Test with pre-computed short_uid parameter
        result3 = get_file_download_url(None, short_uid="MyCustom-Short-UID")
        self.assertEqual(result3, ("https://downloads.files.com/MyCustom-Short-UID", "MyCustom-Short-UID"))

    def test_get_max_download_date(self):
        annex = self.folders[0]
        mod_date = annex.modified().asdatetime().date()
        self.assertEqual(get_max_download_date(annex), mod_date + timedelta(days=90))
        self.assertEqual(get_max_download_date(annex, timedelta(days=0)), mod_date)
        today = date.today()
        self.assertEqual(get_max_download_date(None, timedelta(days=0), today), today)

    def test_create_external_session_not_found(self):
        """Returns _session_not_found_ for a non-existent session id."""
        result = create_external_session(9999)
        self.assertEqual(result, "_session_not_found_")

    def test_create_external_session_no_seal_email(self):
        """Returns _no_seal_email_ when session requires a seal but no email configured."""
        sid, _session = add_files_to_session([], (self.uids[0],), seal="SEAL")
        # seal_email defaults to "" (falsy) — no registry setup needed
        result = create_external_session(sid, esign_root_url="http://test.example.com")
        self.assertEqual(result, "_no_seal_email_")

    def test_create_external_session_no_seal_code(self):
        """Returns _no_seal_code_ when seal email is set but seal code is missing."""
        sid, _session = add_files_to_session([], (self.uids[0],), seal="SEAL")
        api.portal.set_registry_record("imio.esign.seal_email", u"seal@example.com")
        self.addCleanup(api.portal.set_registry_record, "imio.esign.seal_email", u"")
        # seal_code defaults to "" (falsy)
        result = create_external_session(sid, esign_root_url="http://test.example.com")
        self.assertEqual(result, "_no_seal_code_")

    def test_create_external_session_error_response(self):
        """Returns response object and leaves session state unchanged on non-200."""
        signers = [("user1", "user1@sign.com", "User 1", "Position 1")]
        sid, session = add_files_to_session(signers, (self.uids[0],))
        mock_response = Mock()
        mock_response.status_code = 500
        mock_response.text = "Internal Server Error"
        with patch("imio.esign.utils.requests.post", return_value=mock_response):
            with patch("imio.esign.utils.get_auth_token", return_value="test-token"):
                result = create_external_session(sid, esign_root_url="http://test.example.com")
        self.assertIs(result, mock_response)
        self.assertEqual(session["state"], "draft")

    def test_create_external_session_sign_payload(self):
        """Returns response object, sets session state to 'sent', and signData contains signer email."""
        signers = [("user1", "user1@sign.com", "User 1", "Position 1")]
        sid, session = add_files_to_session(signers, (self.uids[0],))
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = '{"message": "OK"}'
        with patch("imio.esign.utils.requests.post", return_value=mock_response) as mock_post:
            with patch("imio.esign.utils.get_auth_token", return_value="test-token"):
                result = create_external_session(sid, esign_root_url="http://test.example.com")
        self.assertIs(result, mock_response)
        self.assertEqual(session["state"], "sent")
        payload = json.loads(mock_post.call_args[1]["data"]["data"])
        self.assertEqual(
            payload,
            {
                u"commonData": {
                    u"imioAppSessionId": u"012345600000",
                    u"vatNumber": None,
                    u"endpointUrl": u"http://nohost/plone/@external_session_feedback",
                    u"sessionName": u"Session 0",
                    u"documentData": [
                        {
                            u"uniqueCode": u"012345600000000__{}".format(self.uids[0]),
                            u"docUuid": get_suid_from_uuid(self.uids[0]),
                            u"filename": u"annex0.pdf",
                        }
                    ],
                },
                u"signData": {u"acroform": True, u"users": [u"user1@sign.com"]},
            },
        )

    def test_create_external_session_seal_payload(self):
        """Seal-only session: sealData contains seal email and code; no signData."""
        sid, _session = add_files_to_session([], (self.uids[0],), seal="SEAL")
        api.portal.set_registry_record("imio.esign.seal_email", u"seal@example.com")
        api.portal.set_registry_record("imio.esign.seal_code", u"PADES_SEAL")
        self.addCleanup(api.portal.set_registry_record, "imio.esign.seal_email", u"")
        self.addCleanup(api.portal.set_registry_record, "imio.esign.seal_code", u"")
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = '{"message": "OK"}'
        with patch("imio.esign.utils.requests.post", return_value=mock_response) as mock_post:
            with patch("imio.esign.utils.get_auth_token", return_value="test-token"):
                create_external_session(sid, esign_root_url="http://test.example.com")
        payload = json.loads(mock_post.call_args[1]["data"]["data"])
        self.assertEqual(
            payload,
            {
                u"commonData": {
                    u"imioAppSessionId": u"012345600000",
                    u"vatNumber": None,
                    u"endpointUrl": u"http://nohost/plone/@external_session_feedback",
                    u"sessionName": u"Session 0",
                    u"documentData": [
                        {
                            u"uniqueCode": u"012345600000000__{}".format(self.uids[0]),
                            u"docUuid": get_suid_from_uuid(self.uids[0]),
                            u"filename": u"annex0.pdf",
                        }
                    ],
                },
                u"sealData": {
                    u"sealCode": u"PADES_SEAL",
                    u"acroform": True,
                    u"users": [u"seal@example.com"],
                    u"watchers": [],
                },
            },
        )

    def test_create_external_session_both_payload(self):
        """Session with signers and seal: both signData and sealData are present."""
        signers = [("user1", "user1@sign.com", "User 1", "Position 1")]
        sid, _session = add_files_to_session(signers, (self.uids[0],), seal="SEAL")
        api.portal.set_registry_record("imio.esign.seal_email", u"seal@example.com")
        api.portal.set_registry_record("imio.esign.seal_code", u"PADES_SEAL")
        set_esign_registry_external_watchers(u"example@imlo.be")  # Also test with one external watcher
        self.addCleanup(api.portal.set_registry_record, "imio.esign.seal_email", u"")
        self.addCleanup(api.portal.set_registry_record, "imio.esign.seal_code", u"")
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = '{"message": "OK"}'
        with patch("imio.esign.utils.requests.post", return_value=mock_response) as mock_post:
            with patch("imio.esign.utils.get_auth_token", return_value="test-token"):
                create_external_session(sid, esign_root_url="http://test.example.com")
        payload = json.loads(mock_post.call_args[1]["data"]["data"])
        self.assertEqual(
            payload,
            {
                u"signData": {
                    u"acroform": True,
                    u"users": [u"user1@sign.com"],
                    u"watchers": [u'example@imlo.be'],
                },
                u"commonData": {
                    u"imioAppSessionId": u"012345600000",
                    u"vatNumber": None,
                    u"endpointUrl": u"http://nohost/plone/@external_session_feedback",
                    u"sessionName": u"Session 0",
                    u"documentData": [
                        {
                            u"uniqueCode": u"012345600000000__{}".format(self.uids[0]),
                            u"docUuid": get_suid_from_uuid(self.uids[0]),
                            u"filename": u"annex0.pdf",
                        }
                    ],
                },
                u"sealData": {
                    u"sealCode": u"PADES_SEAL",
                    u"acroform": True,
                    u"users": [u"seal@example.com"],
                    u"watchers": [u'example@imlo.be'],
                },
            },
        )


# example of annotation content
"""
{
    "numbering": 1,
    "uids": {"3c0528c0ad364641be8b9cbaedbf6620": 0},
    "c_uids": {"f66b3da2d2e947fd81ab65e3e36c039d": ["3c0528c0ad364641be8b9cbaedbf6620"]},
    "sessions": {
        0: {
            "acroform": True,
            "cliend_id": "0123456",
            "discriminators": (),
            "files": [
                {
                    "context_uid": "f66b3da2d2e947fd81ab65e3e36c039d",
                    "scan_id": "012345600000000",
                    "title": u"Annex 0",
                    "uid": "3c0528c0ad364641be8b9cbaedbf6620",
                    "filename": u"annex0.pdf",
                }
            ],
            "last_update": datetime.datetime(2025, 8, 13, 13, 22, 41, 107895),
            "returns": []
            "seal": None,
            "sign_id": None,
            "sign_url": None,
            "signers": [
                {
                    "status": "",
                    "userid": "user1",
                    "email": "user1@sign.com",
                    "fullname": "User 1",
                    "position": "Position 1",
                },
                {
                    "status": "",
                    "userid": "user2",
                    "email": "user2@sign.com",
                    "fullname": "User 2",
                    "position": "Position 2",
                },
            ],
            "state": "draft",
            "title": "my title",
        }
    },
}
"""



# -*- coding: utf-8 -*-

from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from datetime import timedelta
from imio.esign import _tr as _
from imio.esign import API_ROOT_URL
from imio.esign import logger
from imio.esign.audit import audit
from imio.esign.config import get_esign_registry_external_watchers
from imio.esign.config import get_esign_registry_file_url
from imio.esign.config import get_esign_registry_max_session_size
from imio.esign.config import get_esign_registry_seal_code
from imio.esign.config import get_esign_registry_seal_email
from imio.esign.config import get_esign_registry_sign_code
from imio.esign.config import get_esign_registry_vat_number
from imio.esign.interfaces import IContextUidProvider
from imio.helpers.content import uuidToObject
from imio.helpers.transmogrifier import get_correct_id
from imio.helpers.ws import get_auth_token
# from imio.pyutils.system import post_request
from imio.pyutils.utils import shortuid_encode_id
from os import path
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from plone import api
from zope.annotation import IAnnotations
from zope.component import getAdapter

import json
import requests


SESSION_URL = "imio/esign/v1/luxtrust/sessions"


def get_filesize(uid):
    """Get the file size of an annex.

    :param uid: The UID of the annex.
    :return: The file size in bytes, or 0 if not found.
    """
    annex = uuidToObject(uuid=uid, unrestricted=True)
    if not annex:
        logger.error("Annex with UID %s not found.", uid)
        return 0
    return getattr(annex.__parent__, "categorized_elements", {}).get(uid, {}).get("filesize", annex.file.size)


def add_files_to_session(
    signers, files_uids, seal=None, acroform=True, session_id=None, title="", discriminators=(), watchers=(),
    create_session_custom_data=None
):
    """Add files to a session with the given signers.

    :param signers: a list of signers, each is a quartet with userid, email, fullname and position text
    :param files_uids: files uids list
    :param seal: seal or not
    :param acroform: boolean to indicate if signer tag is in files
    :param session_id: session number
    :param title: optional string for session title. If it contains {sign_id} or {session_id} it will be replaced
    :param discriminators: optional list of string discriminators to use for session discrimination
    :param watchers: optional list of external esign session watchers emails (used only when creating a new session)
    :param create_session_custom_data: optional custom dict of custom session data
    :return: session_id, session
    """
    annot = get_session_annotation()
    size = sum(get_filesize(uid) for uid in files_uids)
    if session_id is not None:
        if session_id not in annot["sessions"]:
            logger.error("Session with id %s not found in esign annotations.", session_id)
            session_id = session = None
        else:
            session = annot["sessions"][session_id]
    else:
        session_id, session = discriminate_sessions(signers, seal, acroform, discriminators=discriminators, size=size)
    if not session:
        session_id, session = create_session(
            signers, seal, acroform=acroform, title=title, annot=annot, discriminators=discriminators,
            watchers=watchers,
            create_session_custom_data=create_session_custom_data,
        )
        audit(
            "create_session",
            "session={} signers={}".format(session_id, "|".join([sg[1] for sg in signers]))
        )
    session["size"] = session.get("size", 0) + size
    existing_files = [path.splitext(f["filename"])[0] for f in session["files"]]
    for uid in files_uids:
        annex = uuidToObject(uuid=uid, unrestricted=True)
        context_uid_provider = getAdapter(annex, IContextUidProvider)
        context_uid = context_uid_provider.get_context_uid()
        # update data if adding same file to same session
        if annot['uids'].get(uid, -1) == session_id:
            logger.info('File with UID %s is already in session_id %s and data were updated!', uid, session_id)
            # remove old filename to avoid filename being renamed
            old_filename = path.splitext([fn for fn in session["files"]
                                          if fn['uid'] == uid][0]['filename'])[0]
            existing_files.remove(old_filename)
            remove_files_from_session([uid], remove_empty_session=False)

        filename, ext = path.splitext(annex.file.filename or "no_filename.pdf")
        new_filename = get_correct_id(existing_files, filename)
        session["files"].append(
            PersistentMapping({
                "scan_id": annex.scan_id,
                "filename": new_filename + ext,
                "title": annex.title or "no_title",
                "uid": uid,
                "context_uid": context_uid,
                "status": "",
            })
        )
        existing_files.append(new_filename)
        annot["uids"][uid] = session_id
        annot["c_uids"].setdefault(context_uid, PersistentList()).append(uid)
        audit(
            "add_files_to_session",
            "session={} context={} file={}".format(session_id, context_uid, uid)
        )
    if session["client_id"] is None:
        # FIXME what if scan_id is None ?
        session["client_id"] = session["files"][0]["scan_id"][0:7]
        session["sign_id"] = "{}{:05d}".format(session["client_id"], session_id)
        if u"{sign_id}" in session["title"]:
            session["title"] = session["title"].replace(u"{sign_id}", session["sign_id"])
        if u"{session_id}" in session["title"]:
            session["title"] = session["title"].replace(u"{session_id}", str(session_id))
    session["last_update"] = datetime.now()
    return session_id, session


def create_external_session(session_id, esign_root_url=None):
    """Create a session with the given signers and files.

    :param session_id: internal session id
    :param esign_root_url: the root URL for the e-sign service, if not provided it will use the default API_ROOT_URL
    :return: session information
    """
    session_url = get_esign_session_url(esign_root_url)
    annot = get_session_annotation()
    session = annot["sessions"].get(session_id)
    if not session:
        logger.error("Session with id %s not found.", session_id)
        return "_session_not_found_"
    files = []
    for file_dic in session["files"]:
        uid = file_dic["uid"]
        annex = uuidToObject(uuid=uid, unrestricted=True)
        if not annex:
            logger.error("Annex with UID %s not found.", uid)
            continue
        files.append((file_dic["scan_id"], file_dic["filename"], annex.file.data, uid))
    if not files:
        logger.error("No files found for session %s.", session_id)
        return "_no_files_"
    portal = api.portal.get()  # noqa F841
    if not session["title"]:
        session["title"] = _("Session ${id}", mapping={"id": session_id})
    data_payload = {
        "commonData": {
            "endpointUrl": portal.absolute_url() + "/@external_session_feedback",
            "documentData": [{"filename": filename, "uniqueCode": "{}__{}".format(unique_code, uid),
                              "docUuid": get_suid_from_uuid(uid)}
                             for unique_code, filename, z, uid in files],
            "imioAppSessionId": session["sign_id"],
            "sessionName": session["title"],
        }
    }
    # not mandatory now
    vat_number = get_esign_registry_vat_number(default="BE0000000097")
    data_payload["commonData"]["vatNumber"] = vat_number

    watchers = list(session.get("watchers", []))
    external_watchers = get_esign_registry_external_watchers()
    watchers.extend([ew for ew in external_watchers if ew not in watchers])
    signers = [fdic["email"] for fdic in session["signers"]]
    if signers:
        data_payload["signData"] = {
            "users": list(signers),
            "acroform": session["acroform"],
        }
        sign_code = get_esign_registry_sign_code()
        if sign_code:
            data_payload["signData"]["signCode"] = sign_code
        if watchers:
            data_payload["signData"]["watchers"] = watchers

    if session["seal"]:
        seal_email = get_esign_registry_seal_email()
        if not seal_email:
            logger.error("No seal email configured in registry.")
            return "_no_seal_email_"
        seal_code = get_esign_registry_seal_code()  # PADES_SEAL
        if not seal_code:
            logger.error("No seal code configured in registry.")
            return "_no_seal_code_"
        data_payload["sealData"] = {
            "users": [seal_email],
            # "placeholderName": "SCEAU",  # default
            "acroform": True,
            "watchers": watchers,
            "sealCode": seal_code,
        }

    # files_payload = {filename: file_content for z, filename, file_content, uid in files}
    files_payload = [("files", (filename, file_content)) for z, filename, file_content, uid in files]

    # Headers avec autorisation
    headers = {
        "accept": "application/json",
        "Authorization": "Bearer %s" % get_auth_token(),
    }

    logger.info(data_payload)
    # for future use when pyutils > 1.2.1
    # ret = post_request(
    #     session_url,
    #     headers=headers,
    #     data={"data": json.dumps(data_payload, default=vars)},
    #     files=files_payload,
    #     timeout=10,
    # )
    ret = requests.post(
        session_url,
        headers=headers,
        data={"data": json.dumps(data_payload, default=vars)},
        files=files_payload,
        timeout=10,
    )
    if ret.status_code == 200:
        session["state"] = "sent"
    logger.info("Response: %s", ret.text)
    # {"message":"Request received in the expected format. Session is being created in background."}
    return ret


def create_session(signers, seal=False, acroform=True, title=None, annot=None, discriminators=(), watchers=(),
                   create_session_custom_data=None):
    """Create a session with the given signers and seal.

    :param signers: a list of signers, each is a quartet with userid, email, fullname and position text
    :param seal: seal boolean
    :param acroform: acroform boolean
    :param title: title of the session
    :param annot: esign annotation, if not provided it will be fetched
    :param discriminators: optional list of string discriminators
    :param watchers: optional list of external esign session watchers emails
    :param create_session_custom_data: optional custom dict of custom session data
    :return: session id and session information
    """
    if not annot:
        annot = get_session_annotation()
    sessions = annot.setdefault("sessions", PersistentMapping())
    session_id = annot["numbering"]
    annot["numbering"] += 1

    sessions[session_id] = PersistentMapping({
        "acroform": acroform,
        "client_id": None,
        "discriminators": discriminators,
        "files": PersistentList(),
        "last_update": datetime.now(),
        "seal": seal,
        "sign_id": None,
        "sign_url": None,
        "signers": PersistentList(
            [
                PersistentMapping({"userid": userid, "email": email, "fullname": fullname, "position": position, "status": ""})
                for userid, email, fullname, position in signers
            ]
        ),
        "watchers": PersistentList(watchers),
        "state": "draft",
        "title": title,
        "returns": PersistentList(),
    })
    if create_session_custom_data:
        for k, v in create_session_custom_data.items():
            sessions[session_id][k] = v
    return session_id, sessions[session_id]


def discriminate_sessions(signers, seal, acroform, discriminators=(), annot=None, size=0):
    """Discriminate sessions based on seal value and signers in the same order.

    :param signers: a list of signers, each is a quartet with userid, email, fullname and position text
    :param seal: seal boolean
    :param acroform: boolean value indicating if acroform is used
    :param discriminators: optional list of string discriminators
    :param size: size in bytes of the files to be added to the session
    :param annot: esign annotation, if not provided it will be fetched
    :return: session id and session if found, or (None, None) if no session found
    """
    if not annot:
        annot = get_session_annotation()
    sessions = annot.get("sessions", {})
    max_session_size = get_esign_registry_max_session_size() * 1024**2

    for session_id, session in sessions.items():
        if session["state"] != "draft":
            continue
        if session.get("seal") != seal:
            continue
        if session.get("acroform") != acroform:
            continue
        session_signers = session.get("signers", [])
        if len(signers) != len(session_signers):
            continue
        if size + session.get("size", 0) > max_session_size:
            continue

        if set(discriminators) != set(session.get("discriminators", ())):
            continue

        signers_match = all(
            (userid, email) == (s["userid"], s["email"]) for (userid, email, z, z), s in zip(signers, session_signers)
        )
        if signers_match:
            return session_id, session

    return None, None


def get_esign_session_url(esign_root_url):
    """Get the e-sign root URL."""
    if esign_root_url:
        return "{}/{}".format(esign_root_url, SESSION_URL)
    else:
        return "{}/{}".format(API_ROOT_URL, SESSION_URL)


def get_session_annotation(portal=None):
    """Get the e-sign session annotation."""
    if not portal:
        portal = api.portal.get()
    annotations = IAnnotations(portal)
    if "imio.esign" not in annotations:
        annotations["imio.esign"] = PersistentMapping(
            {
                "numbering": 0,
                "sessions": PersistentMapping(),
                "uids": PersistentMapping(),
                "c_uids": PersistentMapping(),
            }
        )
    return annotations["imio.esign"]


def get_file_info(session_id, file_uid, portal=None, readonly=True):
    """Return informations about a file (uid, title, filename, ...) in a session.

    :param session_id: the session id to return
    :param file_uid: the file UID in the session
    :param portal: portal if necessary to get the session annotation
    :param readonly: return a copy of stored data to avoid modifying it
    """
    session = get_session_info(session_id, portal=portal, readonly=readonly)
    if session:
        for file_info in session['files']:
            if file_info['uid'] == file_uid:
                if readonly:
                    file_info = deepcopy(file_info)
                return file_info


def get_session_info(session_id, portal=None, readonly=True):
    """Return a session info for a given numbering.

    :param session_id: the session id to return
    :param portal: portal if necessary to get the session annotation
    :param readonly: return a copy of stored data to avoid modifying it
    """
    annot = get_session_annotation(portal=portal)
    session = {}
    if session_id in annot['sessions']:
        session = annot['sessions'][session_id]
        if readonly:
            session = deepcopy(session)
    return session


def remove_context_from_session(context_uids):
    """Remove all files from a session that are linked to the given context UIDs.

    :param context_uids: context_uids list
    """
    annot = get_session_annotation()
    c_uids = annot["c_uids"]
    for context_uid in context_uids:
        if context_uid not in c_uids:
            logger.error("Context UID %s not found in session", context_uid)
            continue
        remove_files_from_session(list(c_uids[context_uid]))


def remove_files_from_session(files_uids, remove_empty_session=True):
    """Remove files from their corresponding sessions.

    :param files_uids: list of file UIDs to remove
    :param remove_empty_session: when the last file of a session is removed the session will be removed by default, except when False, the empty session is kept
    """
    annot = get_session_annotation()
    sessions = annot["sessions"]
    uids = annot["uids"]
    c_uids = annot["c_uids"]

    for uid in files_uids:
        session_id = uids.get(uid)
        if session_id is None:
            logger.error("No session found for file UID %s", uid)
            continue
        del uids[uid]
        if session_id not in sessions:
            logger.error("Session %s not found", session_id)
            continue
        session = sessions[session_id]
        session["size"] = max(0, session.get("size", 0) - get_filesize(uid))
        i = 0
        context_uid = None
        for j, dic in enumerate(session["files"]):
            if dic["uid"] == uid:
                i = j
                context_uid = dic["context_uid"]
                break
        else:
            logger.error("File UID %s not found in session %s", uid, session_id)
            continue

        del session["files"][i]
        if not session["files"] and remove_empty_session:
            del sessions[session_id]
        else:
            session["last_update"] = datetime.now()

        if context_uid in c_uids and uid in c_uids[context_uid]:
            c_uids[context_uid].remove(uid)
            if not c_uids[context_uid]:
                del c_uids[context_uid]

        # logger.info("File UID %s removed from session %s", uid, session_id)


def remove_session(session_id):
    """Remove a complete session and all its associated files.

    :param session_id: ID of the session to remove
    """
    annot = get_session_annotation()
    sessions = annot["sessions"]
    uids = annot["uids"]
    c_uids = annot["c_uids"]

    if session_id not in sessions:
        logger.error("Session %s not found", session_id)
        return

    session = sessions[session_id]
    for fdic in session["files"]:
        if fdic["uid"] in uids:
            del uids[fdic["uid"]]
        if fdic["context_uid"] in c_uids and fdic["uid"] in c_uids[fdic["context_uid"]]:
            c_uids[fdic["context_uid"]].remove(fdic["uid"])
            if not c_uids[fdic["context_uid"]]:
                del c_uids[fdic["context_uid"]]

    del sessions[session_id]
    # logger.info("Session %s removed", session_id)


def get_file_download_url(uid, root_url=None, short_uid=None):
    """Get the file download URL for a given file UID.

    :param uid: file UID
    :param root_url: root URL. If not provided, the settings value is used
    :param short_uid: take this short UID instead of computing it
    :return: file download URL, short_uid
    """
    if not root_url:
        root_url = get_esign_registry_file_url()

    if not root_url:
        raise Exception("No root URL provided for file download url.")
    if not short_uid:
        short_uid = get_suid_from_uuid(uid)
    return "{}/{}".format(root_url.strip('/'), short_uid), short_uid


def get_max_download_date(obj, delta=timedelta(days=90), adate=None):
    """Get the maximum download date for e-sign files. Is takes the modification date and adds delta.

    :param obj: content object
    :param delta: timedelta to add to modification date
    :param adate: use date instead of modification date
    :return: maximum download date
    """
    if adate is None:
        adate = obj.modified().asdatetime().date()
    return adate + delta


def get_suid_from_uuid(uid):
    """Get the short UID from a given UUID.

    :param uid: UUID
    :return: short UID
    """
    return shortuid_encode_id(uid, separator="-", block_size=5)


def persistent_to_native(value):
    """Convert persistent object to native object recursively."""
    if isinstance(value, (PersistentMapping, dict)):
        return {k: persistent_to_native(v) for k, v in value.items()}
    elif isinstance(value, (PersistentList, list, tuple)):
        return [persistent_to_native(v) for v in value]
    return value


def get_state_description(state):
    """
    Get a human readable description for a given session state.

                       ┌─────────┐
                       │  draft  │
                       └────┬────┘
                            │
                 (session sent to Paraphéo)
                            │
                            ▼
                       ┌─────────┐
             ┌─────────│  sent   │─────────┐
             │         └────┬────┘         │
             │              │              │
    (error occurred)     (ready)   (signer refused)
             │              │              │
             ▼              ▼              ▼
        ┌─────────┐    ┌─────────┐    ┌─────────┐
        │ errored │    │ to_sign │    │ refused │
        └─────────┘    └────┬────┘    └─────────┘
                            │
                   (documents signed)
                            │
                            ├──────────────┐
                            │              │
            (sent back successfully)  (send back failed)
                            │              │
                            ▼              │
                      ┌──────────┐         │
                      │ returned │         │
                      └─────┬────┘         │
                            │              │
                 (documents received)      │
                            │              │
                            ▼              ▼
                      ┌───────────┐   ┌─────────┐
                      │ finalized │   │ signed  │
                      └───────────┘   └─────────┘
    """
    return {
        'draft': u'The session is getting ready to be sent to Paraphéo by a signing manager.',
        'sent': u'The session has been sent to Paraphéo.',
        'errored': u'The session encountered an error during its processing.',
        'to_sign': u'The session is ready to be signed in Paraphéo.',
        'signed': u'The session is finished but signed documents couldn\'t be sent back to the application.',
        'refused': u'The session has been cancelled because a signer refused a document.',
        'returned': u'The session is finished and signed documents are on the way back to the application.',
        'finalized': u'The session is finished and signed documents have been sent back to the application.',
    }.get(state, "")


def get_sessions_for(context_uid, readonly=True):
    """Returns a list of all sessions involving the provided context_uid"""
    annot = get_session_annotation()
    sessions = OrderedDict()
    seen = set()
    for f_uid in annot["c_uids"].get(context_uid, []):
        session_id = annot["uids"].get(f_uid)
        if session_id is not None and session_id not in seen:
            seen.add(session_id)
            session = annot["sessions"][session_id]
            if readonly:
                session = deepcopy(session)
            sessions[session_id] = session
    return sessions
